"""
Original Algorithm for Shortest path (Dijkstra with a 4-ary heap) was written by François Pacull
<francois.pacull@architecture-performance.fr> under license: MIT, (C) 2022
"""

"""
TODO:
LIST OF ALL THE THINGS WE NEED TO DO TO NOT HAVE TO HAVE nodes 1..n as CENTROIDS. ARBITRARY NUMBERING
- Checks of weather the centroid we are computing path from is a centroid and/or exists in the graph
- Re-write function **network_loading** on the part of loading flows to centroids
"""
cimport cython
from libc.math cimport INFINITY, sin, cos, asin, sqrt, pi
from libc.string cimport memset
from libc.stdlib cimport malloc, free

include 'pq_4ary_heap.pyx'


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)  # turn off bounds-checking for entire function
cpdef void network_loading(
    long classes,
    double[:, :] demand,
    long long [:] pred,
    long long [:] conn,
    double[:, :] link_loads
) noexcept nogil:
    cdef long long i, j, predecessor, connector, node
    cdef long long zones = demand.shape[0]

    # Traditional loading, without cascading
    for i in range(zones):
        node = i

        predecessor = pred[node]
        connector = conn[node]
        while predecessor >= 0:
            for j in range(classes):
                link_loads[connector, j] += demand[i, j]

            connector = conn[predecessor]
            predecessor = pred[predecessor]


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cdef void _copy_skims(
    double[:, :] skim_matrix,  # Skim matrix_procedures computed from one origin to all nodes
    double[:, :] final_skim_matrix
) noexcept nogil:  # Skim matrix_procedures computed for one origin to all other centroids only

    cdef long i, j
    cdef long N = final_skim_matrix.shape[0]
    cdef long skims = final_skim_matrix.shape[1]

    for i in range(N):
        for j in range(skims):
            final_skim_matrix[i, j]=skim_matrix[i, j]


cdef int[:] return_an_int_view(input) noexcept nogil:
    cdef int [:] critical_links_view = input
    return critical_links_view


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cdef void sl_network_loading(
    long long [:, :] selected_links,
    double [:, :] demand,
    long long [:] pred,
    long long [:] conn,
    double [:, :] link_loads,
    double [:, :, :] sl_od_matrix,
    double [:, :, :] sl_link_loading,
    unsigned char [:] has_flow_mask,
    long classes
) noexcept nogil:
    # VARIABLES:
    #   selected_links: 2d memoryview. Each row corresponds to a set of selected links specified by the user
    #   demand:         The input demand matrix for a given origin. The first index corresponds to destination,
    #                   second is the class
    #   pred:           The list of predecessor nodes, i.e. given a node, referencing that node's index in pred
    #                   yields the previous node in the minimum spanning tree
    #   conn:           The list of links which connect predecessor nodes. referencing it by the predecessor yields
    #                   the link it used to connect the two nodes
    # link_loads:       Stores the loading on each link. Equivalent to link_loads in network_loading
    # temp_sl_od_matrix:     Stores the OD matrix for each set of selected links sliced for the given origin
    # The indices are:  set of links, destination, class
    # temp_sl_link_loading:  Stores the loading on the Selected links, and the paths which use the selected links
    #                   The indices are: set of links, link_id, class)
    # has_flow_mask:    An array which acts as a flag for which links were used in retracing a given OD path
    # classes:          the number of subclasses of vehicles for the given TrafficClass
    #
    # Executes regular loading, while keeping track of SL links
    cdef:
        int i, j, k, l, dests = demand.shape[0], xshape = has_flow_mask.shape[0]
        long long predecessor, connection
        bint found
    for j in range(dests):
        memset(&has_flow_mask[0], 0, xshape * sizeof(unsigned char))
        connection = conn[j]
        predecessor = pred[j]

        # Walk the path and mark all used links in the has_flow_mask
        while predecessor >= 0:
            for k in range(classes):
                link_loads[connection, k] += demand[j, k]
            has_flow_mask[connection] = 1
            connection = conn[predecessor]
            predecessor = pred[predecessor]
        # Now iterate through each SL set and see if any of their links were marked
        for i in range(selected_links.shape[0]):
            # We check to see if the given OD path marked any of our selected links
            found = 0
            l = 0
            while l < selected_links.shape[1] and found == 0:
                # Checks to see if the current set of selected links has finished
                # NOTE: -1 is a default value for the selected_links array. It cannot be a link id, if -1 turns up
                # There is either a serious bug, or the program has reached the end of a set of links in SL.
                # This lets us early exit from the loop without needing to iterate through the rest of the array
                # Which just has placeholder values
                if selected_links[i][l] == -1:
                    break
                if has_flow_mask[selected_links[i][l]] != 0:
                    found = 1
                l += 1
            if found == 0:
                continue
            for k in range(classes):
                sl_od_matrix[i, j, k] = demand[j, k]
            connection = conn[j]
            predecessor = pred[j]
            while predecessor >= 0:
                for k in range(classes):
                    sl_link_loading[i, connection, k] += demand[j, k]
                connection = conn[predecessor]
                predecessor = pred[predecessor]


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void put_path_file_on_disk(unsigned int orig,
                                 unsigned int [:] pred,
                                 long long [:] predecessors,
                                 unsigned int [:] conn,
                                 long long [:] connectors,
                                 long long [:] all_nodes,
                                 unsigned int [:] origins_to_write,
                                 unsigned int [:] nodes_to_write) noexcept nogil:
    cdef long long i
    cdef long long k = pred.shape[0]

    for i in range(k):
        origins_to_write[i] = orig
        nodes_to_write[i] = all_nodes[i]
        pred[i] = all_nodes[predecessors[i]]
        conn[i] = connectors[i]


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cdef void blocking_centroid_flows(int action,
                                  long long orig,
                                  long long centroids,
                                  long long [:] fs,
                                  long long [:] temp_b_nodes,
                                  long long [:] real_b_nodes) noexcept nogil:
    cdef long long i

    if action == 1:  # We are unblocking
        for i in range(fs[centroids]):
            temp_b_nodes[i] = real_b_nodes[i]
    else:  # We are blocking:
        for i in range(fs[centroids]):
            temp_b_nodes[i] = orig

        for i in range(fs[orig], fs[orig + 1]):
            temp_b_nodes[i] = real_b_nodes[i]


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)  # turn of bounds-checking for entire function
cdef void skim_single_path(long origin,
                           long nodes,
                           long skims,
                           double[:, :] node_skims,
                           long long [:] pred,
                           long long [:] conn,
                           double[:, :] graph_costs,
                           long long [:] reached_first,
                           long found) noexcept nogil:
    cdef long long i, node, predecessor, connector, j

    # sets all skims to infinity
    for i in range(nodes):
        for j in range(skims):
            node_skims[i, j] = INFINITY

    # Zeroes the intrazonal cost
    for j in range(skims):
        node_skims[origin, j] = 0

    # Cascade skimming
    for i in range(1, found + 1):
        node = reached_first[i]

        # captures how we got to that node
        predecessor = pred[node]
        connector = conn[node]

        for j in range(skims):
            node_skims[node, j] = node_skims[predecessor, j] + graph_costs[connector, j]


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)  # turn of bounds-checking for entire function
cpdef void skim_multiple_fields(long origin,
                                long nodes,
                                long zones,
                                long skims,
                                double[:, :] node_skims,
                                long long [:] pred,
                                long long [:] conn,
                                double[:, :] graph_costs,
                                long long [:] reached_first,
                                long found,
                                double [:, :] final_skims) noexcept nogil:
    cdef long long i, node, predecessor, connector, j

    # sets all skims to infinity
    for i in range(nodes):
        for j in range(skims):
            node_skims[i, j] = INFINITY

    # Zeroes the intrazonal cost
    for j in range(skims):
        node_skims[origin, j] = 0

    # Cascade skimming
    for i in range(1, found + 1):
        node = reached_first[i]

        # captures how we got to that node
        predecessor = pred[node]
        connector = conn[node]

        for j in range(skims):
            node_skims[node, j] = node_skims[predecessor, j] + graph_costs[connector, j]

    for i in range(zones):
        for j in range(skims):
            final_skims[i, j] = node_skims[i, j]


# ######################################################################################################################
########################################################################################################################
# Original Dijkstra implementation by François Pacull, taken from https://github.com/Edsger-dev/priority_queues
# Old Numpy Buffers were replaces with latest memory views interface to allow for the release of the GIL
# Path tracking arrays and skim arrays were also added to it
########################################################################################################################
# ######################################################################################################################


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)  # turn of bounds-checking for entire function
cpdef int path_finding(
    long origin,
    unsigned char [:] destinations,
    long long destination_count,
    double[:] graph_costs,
    long long [:] csr_indices,
    long long [:] graph_fs,
    long long [:] pred,
    long long [:] ids,
    long long [:] connectors,
    long long [:] reached_first
) noexcept nogil:
    cdef unsigned int M = pred.shape[0]
    cdef:
        size_t tail_vert_idx, head_vert_idx, idx  # indices
        DTYPE_t tail_vert_val, head_vert_val  # vertex travel times
        PriorityQueue pqueue  # binary heap
        ElementState vert_state  # vertex state
        size_t origin_vert = <size_t>origin
        ITYPE_t found = 0

    for i in range(M):
        pred[i] = -1
        connectors[i] = -1
        reached_first[i] = -1

    # initialization of the heap elements
    # all nodes have INFINITY key and NOT_IN_HEAP state
    init_heap(&pqueue, <size_t>M)

    # the key is set to zero for the origin vertex,
    # which is inserted into the heap
    insert(&pqueue, origin_vert, 0.0)

    # main loop
    while pqueue.size > 0:
        tail_vert_idx = extract_min(&pqueue)
        reached_first[found] = tail_vert_idx
        found += 1

        if destination_count < 0:
            pass  # early exit is disabled
        elif destination_count > 0 and destinations[tail_vert_idx]:
            destinations[tail_vert_idx] = False
            destination_count = destination_count - 1

        # If we've just found the last destination, we can exit here. No need to explore any more edges
        if destination_count == 0:
            # If we wish to reuse the tree we've constructed in update_path_trace we need to mark the un-scanned
            # nodes as unreachable. The nodes not in the heap (NOT_IN_HEAP) are already -1
            for idx in range(pqueue.length):
                if pqueue.Elements[idx].state == IN_HEAP:
                    pred[idx] = -1
                    connectors[idx] = -1
            break

        tail_vert_val = pqueue.Elements[tail_vert_idx].key

        # loop on outgoing edges
        for idx in range(<size_t>graph_fs[tail_vert_idx], <size_t>graph_fs[tail_vert_idx + 1]):
            head_vert_idx = <size_t>csr_indices[idx]
            vert_state = pqueue.Elements[head_vert_idx].state
            if vert_state != SCANNED:
                head_vert_val = tail_vert_val + graph_costs[idx]
                if head_vert_val == INFINITY:
                    continue
                elif vert_state == NOT_IN_HEAP:
                    insert(&pqueue, head_vert_idx, head_vert_val)
                    pred[head_vert_idx] = tail_vert_idx
                    connectors[head_vert_idx] = ids[idx]
                elif pqueue.Elements[head_vert_idx].key > head_vert_val:
                    decrease_key(&pqueue, head_vert_idx, head_vert_val)
                    pred[head_vert_idx] = tail_vert_idx
                    connectors[head_vert_idx] = ids[idx]

    free_heap(&pqueue)
    return found - 1


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)  # turn of bounds-checking for entire function
cpdef void dfs(long origin,
               long long [:] csr_indices,
               long long [:] graph_fs,
               long long [:] pred) noexcept nogil:

    cdef:
        size_t tail_vert_idx, head_vert_idx  # indices
        unsigned int M = pred.shape[0]
        vector[ITYPE_t] visited
        size_t origin_vert = <size_t>origin

    for i in range(M):
        pred[i] = -1

    pred[origin_vert] = 0
    # initialization of the list of nodes to be analysed
    visited.push_back(origin_vert)

    # main loop
    while not visited.empty():
        tail_vert_idx = visited.back()
        visited.pop_back()

        # loop on outgoing edges
        for idx in range(<size_t>graph_fs[tail_vert_idx], <size_t>graph_fs[tail_vert_idx + 1]):
            head_vert_idx = <size_t>csr_indices[idx]
            if pred[head_vert_idx] < 0:
                pred[head_vert_idx] = tail_vert_idx
                visited.push_back(head_vert_idx)

    visited.clear()


cdef enum Heuristic:
    HAVERSINE
    EQUIRECTANGULAR

HEURISTIC_MAP = {"haversine": HAVERSINE, "equirectangular": EQUIRECTANGULAR}


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cdef inline double haversine_heuristic(double lat1, double lon1, double lat2, double lon2, void* data) noexcept nogil:
    """
    A haversine heuristic written to minimise expensive (trig) operations.

    Arguments:
        **lat1** (:obj:`double`): Latitude of destination
        **lon1** (:obj:`double`): Longitude of destination
        **lat2** (:obj:`double`): Latitude of node to evalutate
        **lon2** (:obj:`double`): Longitude of node to evalutate
        **data** (:obj:`void*`): This void pointer should hold a precomputed cos(lat1) as a double

    Returns the distance between (lat1, lon1) and (lat2, lon2).
    """
    cdef:
        double cos_lat1 = (<double*>data)[0]  # Cython doesn't support c-style derefs, use array notation instead
        double dlat = lat2 - lat1
        double dlon = lon2 - lon1
        double sin_dlat = sin(dlat / 2)
        double sin_dlon = sin(dlon / 2)
        double a = sin_dlat * sin_dlat + cos_lat1 * cos(lat2) * sin_dlon * sin_dlon
    return 2.0 * 6371000.0 * asin(sqrt(a))


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cdef inline double equirectangular_heuristic(
    double lat1,
    double lon1,
    double lat2,
    double lon2,
    void* _data
) noexcept nogil:
    """
    A Equirectangular approximation heuristic, useful for small distances.
    Not admissible for large distances. A* may not return the optimal path with this heuristic.

    Arguments:
        **lat1** (:obj:`double`): Latitude of destination
        **lon1** (:obj:`double`): Longitude of destination
        **lat2** (:obj:`double`): Latitude of node to evaluate
        **lon2** (:obj:`double`): Longitude of node to evaluate
        **data** (:obj:`void*`): Unused void pointer, for compatibility with other heuristics

    Returns the distance between (lat1, lon1) and (lat2, lon2).

    Reference: https://www.movable-type.co.uk/scripts/latlong.html
    """
    cdef:
        double x = (lon2 - lon1) * cos((lat1 + lat2) / 2.0)
        double y = (lat2 - lat1)
    return 6371000.0 * sqrt(x * x + y * y)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void path_finding_a_star(long origin,
                               long destination,
                               double[:] graph_costs,
                               long long [:] csr_indices,
                               long long [:] graph_fs,
                               long long [:] nodes_to_indices,
                               double [:] lats,
                               double [:] lons,
                               long long [:] pred,
                               long long [:] ids,
                               long long [:] connectors,
                               Heuristic heuristic) noexcept nogil:
    """
    Based on the pseudocode presented at https://en.wikipedia.org/wiki/A*_search_algorithm#Pseudocode
    The following variables have been renamed to be consistent with out Dijkstra's implementation
     - openSet: pqueue
     - cameFrom: pred
     - fScore: pqueue.Elements[idx].key, for some idx
    """

    cdef unsigned int M = pred.shape[0]

    cdef:
        size_t current, neighbour, idx  # indices
        DTYPE_t tentative_gScore  # vertex travel times
        PriorityQueue pqueue  # binary heap
        size_t origin_vert = <size_t>origin
        size_t destination_vert = <size_t>destination if destination != -1 else 0
        double *gScore = <double *>malloc(M * sizeof(double))

    cdef:
        double deg2rad = pi / 180.0
        double lat1_rad = lats[destination_vert] * deg2rad
        double lon1_rad = lons[destination_vert] * deg2rad
        double h, cos_lat1 = cos(lat1_rad)
        double (*heur)(double, double, double, double, void*) noexcept nogil
        void* data

    if heuristic == HAVERSINE:
        heur = haversine_heuristic
        data = <void*>&cos_lat1
    else:  # heuristic == EQUIRECTANGULAR:
        heur = equirectangular_heuristic
        data = <void*>NULL

    for i in range(M):
        pred[i] = -1
        connectors[i] = -1
        gScore[i] = INFINITY

    # initialization of the heap elements
    # all nodes have INFINITY key and NOT_IN_HEAP state
    init_heap(&pqueue, <size_t>M)

    # the key is set to zero for the origin vertex,
    # which is inserted into the heap
    insert(&pqueue, origin_vert, 0.0)
    gScore[origin_vert] = 0.0

    # main loop
    while pqueue.size > 0:
        current = extract_min(&pqueue)

        if current == destination_vert:
            break

        # loop on outgoing edges
        for idx in range(<size_t>graph_fs[current], <size_t>graph_fs[current + 1]):
            neighbour = <size_t>csr_indices[idx]

            tentative_gScore = gScore[current] + graph_costs[idx]
            if tentative_gScore < gScore[neighbour]:
                pred[neighbour] = current
                connectors[neighbour] = ids[idx]
                gScore[neighbour] = tentative_gScore

                h = heur(lat1_rad, lon1_rad, lats[neighbour] * deg2rad, lons[neighbour] * deg2rad, data)

                # Unlike Dijkstra's we can remove a node from the heap and rediscover it with a cheaper path
                if pqueue.Elements[neighbour].state != IN_HEAP:
                    insert(&pqueue, neighbour, tentative_gScore + h)
                else:
                    decrease_key(&pqueue, neighbour, tentative_gScore + h)

    free_heap(&pqueue)
    free(gScore)



from libc.math cimport pow
from cython.parallel import prange


def bpr2(congested_times, link_flows, capacity, fftime, alpha, beta, cores):
    cdef int c = cores

    cdef double [:] congested_view = congested_times
    cdef double [:] link_flows_view = link_flows
    cdef double [:] capacity_view = capacity
    cdef double [:] fftime_view = fftime
    cdef double [:] alpha_view = alpha
    cdef double [:] beta_view = beta

    bpr2_cython(congested_view, link_flows_view, capacity_view, fftime_view, alpha_view, beta_view, c)


def delta_bpr2(dbpr2, link_flows, capacity, fftime, alpha, beta, cores):
    cdef int c = cores

    cdef double [:] dbpr2_view = dbpr2
    cdef double [:] link_flows_view = link_flows
    cdef double [:] capacity_view = capacity
    cdef double [:] fftime_view = fftime
    cdef double [:] alpha_view = alpha
    cdef double [:] beta_view = beta

    dbpr2_cython(dbpr2_view, link_flows_view, capacity_view, fftime_view, alpha_view, beta_view, c)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void bpr2_cython(
    double[:] congested_time,
    double[:] link_flows,
    double [:] capacity,
    double [:] fftime,
    double[:] alpha,
    double [:] beta,
    int cores
) noexcept:
    cdef long long i
    cdef long long l = congested_time.shape[0]

    for i in prange(l, nogil=True, num_threads=cores):
        if link_flows[i] > 0:
            if link_flows[i] > capacity[i]:
                congested_time[i] = fftime[i] * (1 + alpha[i] * (
                    pow(link_flows[i] / capacity[i], 2*beta[i])))
            else:
                congested_time[i] = fftime[i] * (1 + alpha[i] * (
                    pow(link_flows[i] / capacity[i], beta[i])))
        else:
            congested_time[i] = fftime[i]


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void dbpr2_cython(
    double[:] deltaresult,
    double[:] link_flows,
    double [:] capacity,
    double [:] fftime,
    double[:] alpha,
    double [:] beta,
    int cores
) noexcept:
    cdef long long i
    cdef long long l = deltaresult.shape[0]

    for i in prange(l, nogil=True, num_threads=cores):
        if link_flows[i] > 0:
            if link_flows[i] > capacity[i]:
                deltaresult[i] = fftime[i] * (alpha[i] * 2 * beta[i] * (
                    pow(link_flows[i] / capacity[i], (2*beta[i])-1))) / (
                    capacity[i])
            else:
                deltaresult[i] = fftime[i] * (alpha[i] * beta[i] * (
                    pow(link_flows[i] / capacity[i], beta[i]-1))) / capacity[i]
        else:
            deltaresult[i] = fftime[i]



import logging
import os
from functools import partial
from pathlib import Path
from tempfile import gettempdir
from typing import List, Dict

import numpy as np
from aequilibrae.paths.AoN import copy_two_dimensions, copy_three_dimensions
from aequilibrae.paths.AoN import linear_combination, linear_combination_skims, aggregate_link_costs
from aequilibrae.paths.AoN import sum_a_times_b_minus_c, linear_combination_1d
from aequilibrae.paths.AoN import triple_linear_combination, triple_linear_combination_skims
from scipy.optimize import root_scalar

from aequilibrae.paths.all_or_nothing import allOrNothing
from aequilibrae.paths.results import AssignmentResults
from aequilibrae.paths.traffic_class import TrafficClass

if False:
    from aequilibrae.paths.traffic_assignment import TrafficAssignment

from aequilibrae.utils.aeq_signal import SIGNAL, simple_progress
from aequilibrae.utils.interface.worker_thread import WorkerThread


class LinearApproximation(WorkerThread):
    equilibration = SIGNAL(object)
    assignment = SIGNAL(object)
    signal = SIGNAL(object)

    def __init__(self, assig_spec, algorithm, project=None) -> None:
        WorkerThread.__init__(self, None)
        self.signal.emit(["set_text", "Linear Approximation"])
        self.logger = project.logger if project else logging.getLogger("aequilibrae")

        self.project_path = project.project_base_path if project else gettempdir()

        self.algorithm = algorithm
        self.rgap_target = assig_spec.rgap_target
        self.max_iter = assig_spec.max_iter
        self.cores = assig_spec.cores
        self.iteration_issue = []
        self.convergence_report = {"iteration": [], "rgap": [], "alpha": [], "warnings": []}
        if algorithm in ["cfw", "bfw"]:
            self.convergence_report["beta0"] = []
            self.convergence_report["beta1"] = []
            self.convergence_report["beta2"] = []

        self.assig = assig_spec  # type: TrafficAssignment

        if None in [
            assig_spec.classes,
            assig_spec.vdf,
            assig_spec.capacity_field,
            assig_spec.time_field,
            assig_spec.vdf_parameters,
        ]:
            all_par = "Traffic classes, VDF, VDF_parameters, capacity field & time_field"
            raise Exception(
                "Parameter missing. Setting the algorithm is the last thing to do "
                f"when assigning. Check if you have all of these: {all_par}"
            )

        self.traffic_classes = assig_spec.classes  # type: List[TrafficClass]
        self.num_classes = len(assig_spec.classes)

        self.cap_field = assig_spec.capacity_field
        self.time_field = assig_spec.time_field
        self.vdf = assig_spec.vdf
        self.vdf_parameters = assig_spec.vdf_parameters
        self.procedure_id = assig_spec.procedure_id

        self.iter = 0
        self.rgap = np.inf
        self.stepsize = 1.0
        self.conjugate_stepsize = 0.0
        self.fw_class_flow = 0
        # rgap can be a bit wiggly, specifying how many times we need to be below target rgap is a quick way to
        # ensure a better result. We might want to demand that the solution is that many consecutive times below.
        self.steps_below_needed_to_terminate = assig_spec.steps_below_needed_to_terminate
        self.steps_below = 0

        # if this is one, we do not have a new direction and will get stuck. Make it 1.
        self.conjugate_direction_max = 0.99999

        # if FW stepsize is zero, we set it to the corresponding MSA stepsize and then need to not make
        # the step direction conjugate to the previous direction.
        self.do_fw_step = False
        self.conjugate_failed = False
        self.do_conjugate_step = False

        # BFW specific stuff
        self.betas = np.array([1.0, 0.0, 0.0])

        # Instantiates the arrays that we will use over and over
        self.capacity = assig_spec.capacity

        # Creates preload vector from preloads
        self.preload = None
        if assig_spec.preloads is not None:
            cols = assig_spec.preloads.columns.difference(["link_id", "direction"])
            self.preload = assig_spec.preloads[cols].sum(axis=1).to_numpy()

        self.free_flow_tt = assig_spec.free_flow_tt
        self.fw_total_flow = assig_spec.total_flow
        self.congested_time = assig_spec.congested_time
        self.vdf_der = np.array(assig_spec.congested_time, copy=True)
        self.congested_value = np.array(assig_spec.congested_time, copy=True)

        self.step_direction = {}  # type: Dict[AssignmentResults]
        self.previous_step_direction = {}  # type: Dict[AssignmentResults]
        self.temp_step_direction_for_copy = {}  # type: Dict[AssignmentResults]

        self.aons = {}

        for c in self.traffic_classes:
            r = AssignmentResults()
            r.prepare(c.graph, c.matrix)
            self.step_direction[c._id] = r

        if self.algorithm in ["cfw", "bfw"]:
            for c in self.traffic_classes:
                for d in [self.step_direction, self.previous_step_direction, self.temp_step_direction_for_copy]:
                    r = AssignmentResults()
                    r.prepare(c.graph, c.matrix)
                    r.compact_link_loads = np.zeros([])
                    r.compact_total_link_loads = np.zeros([])
                    d[c._id] = r

    def calculate_conjugate_stepsize(self):
        self.vdf.apply_derivative(
            self.vdf_der, self.fw_total_flow, self.capacity, self.free_flow_tt, *self.vdf_parameters, self.cores
        )
        numerator = 0.0
        denominator = 0.0
        prev_dir_minus_current_sol = {}
        aon_minus_current_sol = {}
        aon_minus_prev_dir = {}

        for c in self.traffic_classes:
            stp_dir = self.step_direction[c._id]
            prev_dir_minus_current_sol[c._id] = np.sum(stp_dir.link_loads[:, :] - c.results.link_loads[:, :], axis=1)
            aon_minus_current_sol[c._id] = np.sum(c._aon_results.link_loads[:, :] - c.results.link_loads[:, :], axis=1)
            aon_minus_prev_dir[c._id] = np.sum(c._aon_results.link_loads[:, :] - stp_dir.link_loads[:, :], axis=1)

        for c_0 in self.traffic_classes:
            for c_1 in self.traffic_classes:
                numerator += prev_dir_minus_current_sol[c_0._id] * aon_minus_current_sol[c_1._id]
                denominator += prev_dir_minus_current_sol[c_0._id] * aon_minus_prev_dir[c_1._id]

        numerator = np.sum(numerator * self.vdf_der)
        denominator = np.sum(denominator * self.vdf_der)

        alpha = numerator / denominator
        if alpha < 0.0:
            self.conjugate_stepsize = 0.0
        elif alpha > self.conjugate_direction_max:
            self.conjugate_stepsize = self.conjugate_direction_max
        else:
            self.conjugate_stepsize = alpha

        # for reporting, we use a different convention, consistent with BFW: beta_0 corresponds to multiplier for AON;
        # in calculations we follow the conventions of our TRB paper.
        self.betas[0] = 1.0 - self.conjugate_stepsize
        self.betas[1] = self.conjugate_stepsize
        self.betas[2] = 0.0

    def calculate_biconjugate_direction(self):
        self.vdf.apply_derivative(
            self.vdf_der, self.fw_total_flow, self.capacity, self.free_flow_tt, *self.vdf_parameters, self.cores
        )
        mu_numerator = 0.0
        mu_denominator = 0.0
        nu_nom = 0.0
        nu_denom = 0.0

        w_ = {}
        x_ = {}
        y_ = {}
        z_ = {}

        for c in self.traffic_classes:
            sd = self.step_direction[c._id].link_loads[:, :]
            psd = self.previous_step_direction[c._id].link_loads[:, :]
            ll = c.results.link_loads[:, :]

            x_[c._id] = np.sum(sd * self.stepsize + psd * (1.0 - self.stepsize) - ll, axis=1)
            y_[c._id] = np.sum(c._aon_results.link_loads[:, :] - ll, axis=1)
            z_[c._id] = np.sum(sd - ll, axis=1)
            w_[c._id] = np.sum(psd - sd, axis=1)

        for c_0 in self.traffic_classes:
            for c_1 in self.traffic_classes:
                mu_numerator += x_[c_0._id] * y_[c_1._id]
                mu_denominator += x_[c_0._id] * w_[c_1._id]
                nu_nom += z_[c_0._id] * y_[c_1._id]
                nu_denom += z_[c_0._id] * z_[c_1._id]

        mu_numerator = np.sum(mu_numerator * self.vdf_der)
        mu_denominator = np.sum(mu_denominator * self.vdf_der)
        if mu_denominator == 0.0:
            mu = 0.0
        else:
            mu = -mu_numerator / mu_denominator
            mu = max(0.0, mu)

        nu_nom = np.sum(nu_nom * self.vdf_der)
        nu_denom = np.sum(nu_denom * self.vdf_der)
        if nu_denom == 0.0:
            nu = 0.0
        else:
            nu = -(nu_nom / nu_denom) + mu * self.stepsize / (1.0 - self.stepsize)
            nu = max(0.0, nu)

        self.betas[0] = 1.0 / (1.0 + nu + mu)
        self.betas[1] = nu * self.betas[0]
        self.betas[2] = mu * self.betas[0]

    def __calculate_step_direction(self):
        """Calculates step direction depending on the method"""
        sd_flows = []

        # 2nd iteration is a fw step. if the previous step replaced the aggregated
        # solution so far, we need to start anew.
        if self.iter == 2 or self.stepsize == 1.0 or self.do_fw_step or self.algorithm in ["msa", "frank-wolfe"]:
            self.do_fw_step = False
            self.do_conjugate_step = True
            self.conjugate_stepsize = 0.0
            for c in self.traffic_classes:
                aon_res = c._aon_results
                stp_dir_res = self.step_direction[c._id]
                copy_two_dimensions(stp_dir_res.link_loads, aon_res.link_loads, self.cores)
                stp_dir_res.total_flows()
                if c.results.num_skims > 0:
                    copy_three_dimensions(stp_dir_res.skims.matrix_view, aon_res.skims.matrix_view, self.cores)
                sd_flows.append(aon_res.total_link_loads)

                if c._selected_links:
                    aux_res = self.aons[c._id].aux_res
                    for name, idx in c._aon_results._selected_links.items():
                        copy_two_dimensions(
                            self.sl_step_dir_ll[c._id][name]["sdr"],
                            np.sum(aux_res.temp_sl_link_loading, axis=0)[idx, :, :],
                            self.cores,
                        )
                        copy_three_dimensions(
                            self.sl_step_dir_od[c._id][name]["sdr"],
                            np.sum(aux_res.temp_sl_od_matrix, axis=0)[idx, :, :, :],
                            self.cores,
                        )

        # 3rd iteration is cfw. also, if we had to reset direction search we need a cfw step before bfw
        elif (self.iter == 3) or (self.do_conjugate_step) or (self.algorithm == "cfw"):
            self.do_conjugate_step = False
            self.calculate_conjugate_stepsize()
            for c in self.traffic_classes:
                sdr = self.step_direction[c._id]
                previous = self.previous_step_direction[c._id]

                copy_two_dimensions(previous.link_loads, sdr.link_loads, self.cores)
                previous.total_flows()
                if c.results.num_skims > 0:
                    copy_three_dimensions(previous.skims.matrix_view, sdr.skims.matrix_view, self.cores)

                linear_combination(
                    sdr.link_loads, sdr.link_loads, c._aon_results.link_loads, self.conjugate_stepsize, self.cores
                )

                if c.results.num_skims > 0:
                    linear_combination_skims(
                        sdr.skims.matrix_view,
                        sdr.skims.matrix_view,
                        c._aon_results.skims.matrix_view,
                        self.conjugate_stepsize,
                        self.cores,
                    )

                if c._selected_links:
                    aux_res = self.aons[c._id].aux_res
                    for name, idx in c._aon_results._selected_links.items():
                        sl_step_dir_ll = self.sl_step_dir_ll[c._id][name]
                        sl_step_dir_od = self.sl_step_dir_od[c._id][name]

                        copy_two_dimensions(
                            sl_step_dir_ll["prev_sdr"],
                            sl_step_dir_ll["sdr"],
                            self.cores,
                        )
                        copy_three_dimensions(
                            sl_step_dir_od["prev_sdr"],
                            sl_step_dir_od["sdr"],
                            self.cores,
                        )

                        linear_combination(
                            sl_step_dir_ll["sdr"],
                            sl_step_dir_ll["sdr"],
                            np.sum(aux_res.temp_sl_link_loading, axis=0)[idx, :, :],
                            self.conjugate_stepsize,
                            self.cores,
                        )

                        linear_combination_skims(
                            sl_step_dir_od["sdr"],
                            sl_step_dir_od["sdr"],
                            np.sum(aux_res.temp_sl_od_matrix, axis=0)[idx, :, :, :],
                            self.conjugate_stepsize,
                            self.cores,
                        )

                sdr.total_flows()
                sd_flows.append(sdr.total_link_loads)
        # biconjugate
        else:
            self.calculate_biconjugate_direction()
            # deep copy because we overwrite step_direction but need it on next iteration
            for c in self.traffic_classes:
                ppst = self.temp_step_direction_for_copy[c._id]  # type: AssignmentResults
                prev_stp_dir = self.previous_step_direction[c._id]  # type: AssignmentResults
                stp_dir = self.step_direction[c._id]  # type: AssignmentResults

                copy_two_dimensions(ppst.link_loads, stp_dir.link_loads, self.cores)
                ppst.total_flows()
                if c.results.num_skims > 0:
                    copy_three_dimensions(ppst.skims.matrix_view, stp_dir.skims.matrix_view, self.cores)

                triple_linear_combination(
                    stp_dir.link_loads,
                    c._aon_results.link_loads,
                    stp_dir.link_loads,
                    prev_stp_dir.link_loads,
                    self.betas,
                    self.cores,
                )

                stp_dir.total_flows()
                if c.results.num_skims > 0:
                    triple_linear_combination_skims(
                        stp_dir.skims.matrix_view,
                        c._aon_results.skims.matrix_view,
                        stp_dir.skims.matrix_view,
                        prev_stp_dir.skims.matrix_view,
                        self.betas,
                        self.cores,
                    )

                if c._selected_links:
                    aux_res = self.aons[c._id].aux_res
                    for name, idx in c._aon_results._selected_links.items():
                        sl_step_dir_ll = self.sl_step_dir_ll[c._id][name]
                        sl_step_dir_od = self.sl_step_dir_od[c._id][name]
                        copy_two_dimensions(
                            sl_step_dir_ll["temp_prev_sdr"],
                            sl_step_dir_ll["sdr"],
                            self.cores,
                        )
                        copy_three_dimensions(
                            sl_step_dir_od["temp_prev_sdr"],
                            sl_step_dir_od["sdr"],
                            self.cores,
                        )

                        triple_linear_combination(
                            sl_step_dir_ll["sdr"],
                            np.sum(aux_res.temp_sl_link_loading, axis=0)[idx, :, :],
                            sl_step_dir_ll["sdr"],
                            sl_step_dir_ll["prev_sdr"],
                            self.betas,
                            self.cores,
                        )

                        triple_linear_combination_skims(
                            sl_step_dir_od["sdr"],
                            np.sum(aux_res.temp_sl_od_matrix, axis=0)[idx, :, :, :],
                            sl_step_dir_od["sdr"],
                            sl_step_dir_od["prev_sdr"],
                            self.betas,
                            self.cores,
                        )

                        copy_two_dimensions(
                            sl_step_dir_ll["prev_sdr"],
                            sl_step_dir_ll["temp_prev_sdr"],
                            self.cores,
                        )
                        copy_three_dimensions(
                            sl_step_dir_od["prev_sdr"],
                            sl_step_dir_od["temp_prev_sdr"],
                            self.cores,
                        )

                sd_flows.append(np.sum(stp_dir.link_loads, axis=1))

                copy_two_dimensions(prev_stp_dir.link_loads, ppst.link_loads, self.cores)
                prev_stp_dir.total_flows()
                if c.results.num_skims > 0:
                    copy_three_dimensions(prev_stp_dir.skims.matrix_view, ppst.skims.matrix_view, self.cores)

        self.step_direction_flow = np.sum(sd_flows, axis=0)

    def __maybe_create_path_file_directories(self):
        path_base_dir = os.path.join(self.project_path, "path_files", self.procedure_id)
        for c in self.traffic_classes:
            if c._aon_results.save_path_file:
                c._aon_results.path_file_dir = os.path.join(
                    path_base_dir, f"iter{self.iter}", f"path_c{c.mode}_{c._id}"
                )
                Path(c._aon_results.path_file_dir).mkdir(parents=True, exist_ok=True)
                if self.iter == 1:  # save simplified graph correspondences, this could change after assignment
                    c.graph.save_compressed_correspondence(path_base_dir, c.mode, c._id)

    def doWork(self):
        self.execute()

    def execute(self):  # noqa: C901
        # We build the fixed cost field

        self.sl_step_dir_ll = {}
        self.sl_step_dir_od = {}

        for c in self.traffic_classes:
            # Copying select link dictionary that maps name to its relevant matrices into the class' results
            c._aon_results._selected_links = c._selected_links
            c.results._selected_links = c._selected_links

            link_loads_step_dir_shape = (
                c.graph.compact_num_links,
                c.results.classes["number"],
            )

            od_step_dir_shape = (
                c.graph.num_zones,
                c.graph.num_zones,
                c.results.classes["number"],
            )

            self.sl_step_dir_ll[c._id] = {}
            self.sl_step_dir_od[c._id] = {}
            for name in c._selected_links.keys():
                self.sl_step_dir_ll[c._id][name] = {
                    "sdr": np.zeros(link_loads_step_dir_shape, dtype=c.graph.default_types("float")),
                    "prev_sdr": np.zeros(link_loads_step_dir_shape, dtype=c.graph.default_types("float")),
                    "temp_prev_sdr": np.zeros(link_loads_step_dir_shape, dtype=c.graph.default_types("float")),
                }

                self.sl_step_dir_od[c._id][name] = {
                    "sdr": np.zeros(od_step_dir_shape, dtype=c.graph.default_types("float")),
                    "prev_sdr": np.zeros(od_step_dir_shape, dtype=c.graph.default_types("float")),
                    "temp_prev_sdr": np.zeros(od_step_dir_shape, dtype=c.graph.default_types("float")),
                }

            # Sizes the temporary objects used for the results
            c.results.prepare(c.graph, c.matrix)
            c._aon_results.prepare(c.graph, c.matrix)
            c.results.reset()

            # Prepares the fixed cost to be used
            if c.fixed_cost_field:
                # divide fixed cost by volume-dependent prefactor (vot) such that we don't have to do it for
                # each occurrence in the objective function. TODO: Need to think about cost skims here, we do
                # not want this there I think
                v = c.graph.graph[c.fixed_cost_field].values[:]
                c.fixed_cost[c.graph.graph.__supernet_id__] = v * c.fc_multiplier / c.vot
                c.fixed_cost[np.isnan(c.fixed_cost)] = 0

            # TODO: Review how to eliminate this. It looks unnecessary
            # Just need to create some arrays for cost
            c.graph.set_graph(self.time_field)

            self.aons[c._id] = allOrNothing(c._id, c.matrix, c.graph, c._aon_results)

        self.logger.info(f"{self.algorithm} Assignment STATS")
        self.logger.info("Iteration, RelativeGap, stepsize")

        msg = "Equilibrium Assignment"
        for self.iter in simple_progress(range(1, self.max_iter + 1), self.signal, msg):  # noqa: B020
            self.iteration_issue = []

            aon_flows = []

            self.__maybe_create_path_file_directories()

            for c in self.traffic_classes:  # type: TrafficClass
                msg = f"All-or-Nothing - Traffic Class: {c._id}"
                self.signal.emit(["set_text", msg])
                # cost = c.fixed_cost / c.vot + self.congested_time #  now only once
                cost = c.fixed_cost + self.congested_time
                aggregate_link_costs(cost, c.graph.compact_cost, c.results.crosswalk)

                aon = self.aons[c._id]  # This is a new object every iteration, with new aux_res
                self.signal.emit(["refresh"])
                self.signal.emit(["reset"])
                aon.signal = self.signal

                aon.execute()
                c._aon_results.link_loads *= c.pce
                c._aon_results.total_flows()
                aon_flows.append(c._aon_results.total_link_loads)

            self.aon_total_flow = np.sum(aon_flows, axis=0)

            flows = []
            if self.iter == 1:
                for c in self.traffic_classes:
                    copy_two_dimensions(c.results.link_loads, c._aon_results.link_loads, self.cores)
                    c.results.total_flows()
                    if c.results.num_skims > 0:
                        copy_three_dimensions(c.results.skims.matrix_view, c._aon_results.skims.matrix_view, self.cores)

                    if c._selected_links:
                        for name, idx in c._aon_results._selected_links.items():
                            # Copy the temporary results into the final od matrix, referenced by link_set name
                            # The temp has an index associated with the link_set name
                            copy_three_dimensions(
                                c.results.select_link_od.matrix[name],  # matrix being written into
                                np.sum(self.aons[c._id].aux_res.temp_sl_od_matrix, axis=0)[
                                    idx, :, :, :
                                ],  # results after the iteration
                                self.cores,  # core count
                            )
                            copy_two_dimensions(
                                c.results.select_link_loading[name],  # ouput matrix
                                np.sum(self.aons[c._id].aux_res.temp_sl_link_loading, axis=0)[idx, :, :],  # matrix 1
                                self.cores,  # core count
                            )
                    flows.append(c.results.total_link_loads)

            else:
                self.__calculate_step_direction()
                self.calculate_stepsize()
                for c in self.traffic_classes:
                    stp_dir = self.step_direction[c._id]

                    cls_res = c.results

                    linear_combination(
                        cls_res.link_loads, stp_dir.link_loads, cls_res.link_loads, self.stepsize, self.cores
                    )

                    if cls_res.num_skims > 0:
                        linear_combination_skims(
                            cls_res.skims.matrix_view,
                            stp_dir.skims.matrix_view,
                            cls_res.skims.matrix_view,
                            self.stepsize,
                            self.cores,
                        )

                    if c._selected_links:
                        for name, idx in c._aon_results._selected_links.items():
                            # Copy the temporary results into the final od matrix, referenced by link_set name
                            # The temp flows have an index associated with the link_set name
                            linear_combination_skims(
                                cls_res.select_link_od.matrix[name],  # output matrix
                                self.sl_step_dir_od[c._id][name]["sdr"],
                                cls_res.select_link_od.matrix[name],  # matrix 2 (previous iteration)
                                self.stepsize,  # stepsize
                                self.cores,  # core count
                            )

                            linear_combination(
                                cls_res.select_link_loading[name],  # output matrix
                                self.sl_step_dir_ll[c._id][name]["sdr"],
                                cls_res.select_link_loading[name],  # matrix 2 (previous iteration)
                                self.stepsize,  # stepsize
                                self.cores,  # core count
                            )

                    cls_res.total_flows()
                    flows.append(cls_res.total_link_loads)

            self.fw_total_flow = np.sum(flows, axis=0)
            if self.preload is not None:
                self.fw_total_flow += self.preload

            if self.algorithm == "all-or-nothing":
                break

            # Check convergence
            # This needs to be done with the current costs, and not the future ones
            converged = self.check_convergence() if self.iter > 1 else False
            self.vdf.apply_vdf(
                self.congested_time,
                self.fw_total_flow,
                self.capacity,
                self.free_flow_tt,
                *self.vdf_parameters,
                self.cores,
            )

            for c in self.traffic_classes:
                if self.time_field in c.graph.skim_fields:
                    k = c.graph.skim_fields.index(self.time_field)
                    aggregate_link_costs(self.congested_time[:], c.graph.compact_skims[:, k], c.results.crosswalk)

            self.convergence_report["iteration"].append(self.iter)
            self.convergence_report["rgap"].append(self.rgap)
            self.convergence_report["warnings"].append("; ".join(self.iteration_issue))
            self.convergence_report["alpha"].append(self.stepsize)

            if self.algorithm in ["cfw", "bfw"]:
                self.convergence_report["beta0"].append(self.betas[0])
                self.convergence_report["beta1"].append(self.betas[1])
                self.convergence_report["beta2"].append(self.betas[2])

            self.logger.info(f"{self.iter},{self.rgap},{self.stepsize}")
            if converged:
                self.steps_below += 1
                if self.steps_below >= self.steps_below_needed_to_terminate:
                    break
            else:
                self.steps_below = 0

            if self.iter < self.max_iter:
                for c in self.traffic_classes:
                    c._aon_results.reset()
                    if self.time_field not in c.graph.skim_fields:
                        continue
                    idx = c.graph.skim_fields.index(self.time_field)
                    c.graph.skims[:, idx] = self.congested_time[:]

            msg = f"Equilibrium Assignment - Iteration: {self.iter}/{self.max_iter} - RGap: {self.rgap:.6}"
            self.signal.emit(["set_text", msg])

        for c in self.traffic_classes:
            c.results.link_loads /= c.pce
            c.results.total_flows()
            c.congested_time = self.congested_time

        if (self.rgap > self.rgap_target) and (self.algorithm != "all-or-nothing"):
            self.logger.error(f"Desired RGap of {self.rgap_target} was NOT reached")
        self.logger.info(f"{self.algorithm} Assignment finished. {self.iter} iterations and {self.rgap} final gap")
        self.signal.emit(["finished"])

    def __derivative_of_objective_stepsize_dependent(self, stepsize, const_term):
        """The stepsize-dependent part of the derivative of the objective function. If fixed costs are defined,
        the corresponding contribution needs to be passed in"""
        x = np.zeros_like(self.fw_total_flow)
        linear_combination_1d(x, self.step_direction_flow, self.fw_total_flow, stepsize, self.cores)
        # x = self.fw_total_flow + stepsize * (self.step_direction_flow - self.fw_total_flow)
        self.vdf.apply_vdf(self.congested_value, x, self.capacity, self.free_flow_tt, *self.vdf_parameters, self.cores)
        link_cost_term = sum_a_times_b_minus_c(
            self.congested_value, self.step_direction_flow, self.fw_total_flow, self.cores
        )
        return link_cost_term + const_term

    def __derivative_of_objective_stepsize_independent(self):
        """The part of the derivative of the objective function that does not dependent on stepsize. Non-zero
        only for fixed cost contributions."""
        class_specific_term = 0.0
        for c in self.traffic_classes:
            # fixed cost is scaled by vot
            class_link_costs = sum_a_times_b_minus_c(
                c.fixed_cost, self.step_direction[c._id].link_loads[:, 0], c.results.link_loads[:, 0], self.cores
            )
            class_specific_term += class_link_costs
        return class_specific_term

    def calculate_stepsize(self):
        """Calculate optimal stepsize in descent direction"""
        if self.algorithm == "msa":
            self.stepsize = 1.0 / self.iter
            return

        class_specific_term = self.__derivative_of_objective_stepsize_independent()
        derivative_of_objective = partial(
            self.__derivative_of_objective_stepsize_dependent, const_term=class_specific_term
        )

        x_tol = max(min(1e-6, self.rgap * 1e-5), 1e-12)

        try:
            min_res = root_scalar(derivative_of_objective, bracket=[0, 1], xtol=x_tol)
            self.stepsize = min_res.root
            if not min_res.converged:
                self.logger.warning("Descent direction stepsize finder has not converged")

            self.conjugate_failed = False

        except ValueError as e:
            # We can have iterations where the objective function is not *strictly* convex, but the scipy method cannot deal
            # with this. Stepsize is then either given by 1 or 0, depending on where the objective function is smaller.
            # However, using zero would mean the overall solution would not get updated, and therefore we assert the stepsize
            # in order to add a small fraction of the AoN. A heuristic value equal to the corresponding MSA step size
            # seems to work well in practice.
            if self.algorithm == "bfw":
                self.betas.fill(-1)
            if derivative_of_objective(0.0) < derivative_of_objective(1.0):
                if self.algorithm == "frank-wolfe" or self.conjugate_failed:
                    tiny_step = 1e-2 / self.iter  # use a fraction of the MSA stepsize. We observe that using 1e-4
                    # works well in practice, however for a large number of iterations this might be too much so
                    # use this heuristic instead.
                    self.logger.warning(f"# Alert: Adding {tiny_step} as step size to make it non-zero. {e.args}")
                    self.stepsize = tiny_step
                else:
                    self.stepsize = 0.0
                    # need to reset conjugate / bi-conjugate direction search
                    self.do_fw_step = True
                    self.conjugate_failed = True
                    self.iteration_issue.append("Found bad conjugate direction step. Performing FW search. {e.args}")
                    # By doing it recursively, we avoid doing the same AoN again
                    self.__calculate_step_direction()
                    self.calculate_stepsize()

            else:
                # Do we want to keep some of the old solution, or just throw away everything?
                self.stepsize = 1.0

        assert 0 <= self.stepsize <= 1.0

    def check_convergence(self):
        """Calculate relative gap and return ``True`` if it is smaller than desired precision"""
        aon_cost = np.sum(self.congested_time * self.aon_total_flow)
        current_cost = np.sum(self.congested_time * self.fw_total_flow)
        self.rgap = abs(current_cost - aon_cost) / current_cost
        if self.rgap_target >= self.rgap:
            return True
        return False



from copy import deepcopy
from os.path import join, realpath
from typing import Union, Dict
import warnings

import geopandas as gpd
import pandas as pd
import shapely.wkb
from shapely.geometry import Point, Polygon, LineString, MultiLineString
from shapely import union_all

from aequilibrae.project.basic_table import BasicTable
from aequilibrae.project.data_loader import DataLoader
from aequilibrae.project.network.connector_creation import connector_creation
from aequilibrae.project.project_creation import run_queries_from_sql_file
from aequilibrae.project.table_loader import TableLoader
from aequilibrae.project.zone import Zone
from aequilibrae.utils.aeq_signal import SIGNAL, simple_progress
from aequilibrae.utils.db_utils import commit_and_close
from aequilibrae.utils.geo_index import GeoIndex


class Zoning(BasicTable):
    """
    Access to the API resources to manipulate the 'zones' table in the project

    .. code-block:: python

        >>> project = create_example(project_path, "coquimbo")

        >>> zoning = project.zoning

        >>> zone_downtown = zoning.get(1)
        >>> zone_downtown.population = 637
        >>> zone_downtown.employment = 10039
        >>> zone_downtown.save()

        # We can also add one more field to the table
        >>> fields = zoning.fields
        >>> fields.add('parking_spots', 'Total licensed parking spots', 'INTEGER')
    """

    def __init__(self, network):
        super().__init__(network.project)
        self.__items: Dict[int, Zone] = {}
        self.network = network
        self.__table_type__ = "zones"
        self.__fields = []
        self.__geo_index = GeoIndex()
        if self.__has_zoning():
            self.__load()

    def new(self, zone_id: int) -> Zone:
        """Creates a new zone

        :Returns:
            **zone** (:obj:`Zone`): A new zone object populated only with ``zone_id`` (but not saved in the model yet)
        """

        if zone_id in self.__items:
            raise Exception(f"Zone ID {zone_id} already exists")

        data = {key: None for key in self.__fields}
        data["zone_id"] = zone_id

        self.project.logger.info(f"Zone with id {zone_id} was created")
        return self.__create_return_zone(data)

    def create_zoning_layer(self):
        """Creates the 'zones' table for project files that did not previously contain it"""

        if not self.__has_zoning():
            qry_file = join(realpath(__file__), "database_specification", "tables", "zones.sql")
            with self.network.project.db_connection as conn:
                run_queries_from_sql_file(conn, self.project.logger, qry_file)
            self.__load()
        else:
            self.project.warning("zones table already exists. Nothing was done", Warning)

    def coverage(self) -> Polygon:
        """Returns a single polygon for the entire zoning coverage

        :Returns:
            **model coverage** (:obj:`Polygon`): Shapely (Multi)polygon of the zoning system.
        """
        with self.network.project.db_connection as conn:
            dt = conn.execute('Select ST_asBinary("geometry") from zones;').fetchall()
        polygons = [shapely.wkb.loads(x[0]) for x in dt]
        return union_all(polygons)

    def get(self, zone_id: str) -> Zone:
        """Get a zone from the model by its ``zone_id``"""
        if zone_id not in self.__items:
            raise ValueError(f"Zone {zone_id} does not exist in the model")
        return self.__items[zone_id]

    def all_zones(self) -> dict:
        """Returns a dictionary with all Zone objects available in the model, using ``zone_id`` as key"""
        return self.__items

    def save(self):
        for item in self.__items.values():
            item.save()

    def add_centroids(self, robust=True):
        """Adds automatic centroids to the network file. It adds centroids to all zones that do not have one
        Centroid is added to the geographic centroid of the zone.

        :Arguments:
            **robust** (:obj:`bool`, *Optional*): Moves the centroid location around to avoid node conflict.
            Defaults to ``True``.
        """
        i = 0
        with commit_and_close(self.project.path_to_file, spatial=True) as conn:
            existing_centroids = pd.read_sql("SELECT node_id from Nodes where is_centroid = 1", conn).node_id.to_numpy()
            for zone_id in simple_progress(self.__items.keys(), SIGNAL(object), "Connecting zones"):
                if zone_id in existing_centroids:
                    continue
                zone = self.__items[zone_id]
                zone.add_centroid(zone.geometry.centroid, robust)
                i += 1
        if i > 0:
            self.project.logger.info(f"{i} new centroids added to the network")
        else:
            self.project.logger.info("No new centroids added to the network")

    def connect_mode(self, mode_id: str, link_types="", connectors=1, limit_to_zone=True):
        """Adds centroid connectors for the desired mode to the network file

        Centroid connectors are created by connecting each zone centroid to one or more nodes selected from
        all those that satisfy the mode and link_types criteria and are inside the zone.

        The selection of the nodes that will be connected is done simply by searching for the node closest to each
        zone centroid, or the N closest nodes to the centroid.

        If fewer candidates than required connectors are found, all candidates are connected.

        CENTROIDS THAT ARE CURRENTLY CONNECTED ARE SKIPPED ALTOGETHER

        :Arguments:
            **mode_id** (:obj:`str`): Mode ID we are trying to connect

            **link_types** (:obj:`str`, *Optional*): String with all the link type IDs that can be considered.
            eg: yCdR. Defaults to ALL link types

            **connectors** (:obj:`int`, *Optional*): Number of connectors to add. Defaults to 1

            **limit_to_zone** (:obj:`bool`): Limits the search for nodes inside the zone. Defaults to ``True``.
        """

        proj_nodes = self.project.network.nodes.data
        links = self.project.network.links

        centroids = proj_nodes.reset_index().query("is_centroid == 1", engine="python").node_id.to_numpy()
        link_data = links.data
        centroid_conn = link_data.query("a_node in @centroids and modes.str.contains(@mode_id)", engine="python")
        connected_centroids = centroid_conn.a_node.to_numpy()

        with (
            commit_and_close(self.project.path_to_file, spatial=True) as conn,
            warnings.catch_warnings(),
        ):
            warnings.filterwarnings("ignore", category=UserWarning, module="geopandas")
            zones_todo = [x for x in self.__items.keys() if x not in connected_centroids]
            for zone_id in simple_progress(zones_todo, SIGNAL(object), "Connecting zones"):
                if zone_id not in centroids:
                    self.project.logger.warning(f"Centroid for zone {zone_id} does not exist. Please create it first.")
                    continue

                zone = self.__items[zone_id]
                area = zone.geometry if limit_to_zone else None
                connector_creation(
                    zone_id=zone_id,
                    mode_id=mode_id,
                    link_types=link_types,
                    connectors=connectors,
                    proj_nodes=proj_nodes,
                    proj_links=link_data,
                    network=self.project.network,
                    conn=conn,
                    delimiting_area=area,
                )

    def get_closest_zone(self, geometry: Union[Point, LineString, MultiLineString]) -> int:
        """Returns the zone in which the given geometry is located.

        If the geometry is not fully enclosed by any zone, the zone closest to
        the geometry is returned

        :Arguments:
            **geometry** (:obj:`Point` or :obj:`LineString`): A Shapely geometry object

        :Returns:
            **zone_id** (:obj:`int`): ID of the zone applicable to the point provided
        """

        nearest = self.__geo_index.nearest(geometry, 10)
        dists = {}
        for zone_id in nearest:
            geo = self.__items[zone_id].geometry
            if geo.contains(geometry):
                return zone_id
            dists[geo.distance(geometry)] = zone_id
        return dists[min(dists.keys())]

    def refresh_geo_index(self):
        self.__geo_index.reset()
        for zone_id, zone in self.__items.items():
            self.__geo_index.insert(feature_id=zone_id, geometry=zone.geometry)

    def __has_zoning(self):
        with self.network.project.db_connection as conn:
            dt = conn.execute("SELECT name FROM sqlite_master WHERE type='table';").fetchall()
        return any("zone" in x[0].lower() for x in dt)

    def __load(self):
        tl = TableLoader()
        with self.network.project.db_connection as conn:
            zones_list = tl.load_table(conn, "zones")
        self.__fields = deepcopy(tl.fields)

        existing_list = [zn["zone_id"] for zn in zones_list]
        if zones_list:
            self.__properties = list(zones_list[0].keys())
        for zn in zones_list:
            if zn["zone_id"] not in self.__items:
                self.__items[zn["zone_id"]] = Zone(zn, self)

        to_del = [key for key in self.__items.keys() if key not in existing_list]
        for key in to_del:
            del self.__items[key]
        self.refresh_geo_index()

    def _remove_zone(self, zone_id: int):
        del self.__items[zone_id]

    def __create_return_zone(self, data):
        zone = Zone(data, self)
        self.__items[zone.zone_id] = zone
        return zone

    @property
    def data(self) -> gpd.GeoDataFrame:
        """Returns all zones data as a Pandas DataFrame

        :Returns:
            **table** (:obj:`GeoDataFrame`): GeoPandas GeoDataFrame with all the nodes
        """
        dl = DataLoader(self.project.path_to_file, "zones")
        return dl.load_table()



# distutils: language = c++

from libcpp.vector cimport vector

import numpy as np
cimport numpy as np
np.import_array()


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)  # turn of bounds-checking for entire function
cpdef void save_path_file(long origin_index,
                          long num_links,
                          long zones,
                          long long [:] pred,
                          long long [:] conn,
                          str path_file,
                          str index_file,
                          bint write_feather) noexcept:

    cdef long long node, predecessor, connector
    cdef vector[long long] path_data
    # could make this an ndarray and not do the conversion, we know the size of the index array is zones
    cdef vector[long long] size_of_path_arrays

    cdef np.npy_intp dims[1]
    cdef np.ndarray[np.longlong_t, ndim=1] numpy_array
    cdef np.npy_intp dims_ind[1]
    cdef np.ndarray[np.longlong_t, ndim=1] numpy_array_ind

    with nogil:
        for node in range(zones):
            predecessor = pred[node]
            # need to check if disconnected, also makes sure o==d is not included
            if predecessor == -1:
                size_of_path_arrays.push_back(<np.longlong_t> path_data.size())  # need to store index here
                continue
            connector = conn[node]
            path_data.push_back(connector)
            while predecessor != -1:
                connector = conn[predecessor]  # connector has to be looked up BEFORE predecessor update
                predecessor = pred[predecessor]
                if (predecessor != -1) and (connector != -1):
                    path_data.push_back(connector)

            size_of_path_arrays.push_back(<np.longlong_t> path_data.size())

    # get a view on data underlying vector, then as numpy array. avoids copying.
    dims[0] = <np.npy_intp> (path_data.size())
    dims_ind[0] = <np.npy_intp> (size_of_path_arrays.size())
    numpy_array = np.PyArray_SimpleNewFromData(1, dims, np.NPY_LONGLONG, &path_data[0])
    numpy_array_ind = np.PyArray_SimpleNewFromData(1, dims_ind, np.NPY_LONGLONG, &size_of_path_arrays[0])

    table1 = pd.DataFrame({"data": numpy_array})
    table2 = pd.DataFrame({"data": numpy_array_ind})

    if write_feather:
        table1.to_feather(path_file)
        table2.to_feather(index_file)
    else:
        table1.to_parquet(path_file)
        table2.to_parquet(index_file)



from libc.math cimport pow
from cython.parallel import prange


def inrets(congested_times, link_flows, capacity, fftime, alpha, beta, cores):
    cdef int c = cores

    cdef double [:] congested_view = congested_times
    cdef double [:] link_flows_view = link_flows
    cdef double [:] capacity_view = capacity
    cdef double [:] fftime_view = fftime
    cdef double [:] alpha_view = alpha
    cdef double [:] beta_view = beta

    inrets_cython(congested_view, link_flows_view, capacity_view, fftime_view, alpha_view, beta_view, c)


def delta_inrets(dbpr, link_flows, capacity, fftime, alpha, beta, cores):
    cdef int c = cores

    cdef double [:] dbpr_view = dbpr
    cdef double [:] link_flows_view = link_flows
    cdef double [:] capacity_view = capacity
    cdef double [:] fftime_view = fftime
    cdef double [:] alpha_view = alpha
    cdef double [:] beta_view = beta

    dinrets_cython(dbpr_view, link_flows_view, capacity_view, fftime_view, alpha_view, beta_view, c)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void inrets_cython(
    double[:] congested_time,
    double[:] link_flows,
    double [:] capacity,
    double [:] fftime,
    double[:] alpha,
    double [:] beta,
    int cores
) noexcept:
    cdef long long i
    cdef long long l = congested_time.shape[0]

    for i in prange(l, nogil=True, num_threads=cores):
        if link_flows[i] > 0:
            if link_flows[i] > capacity[i]:
                congested_time[i] = fftime[i] * (
                    (1.1 - alpha[i])/0.1) * (
                    pow(link_flows[i] / capacity[i], 2))
            else:
                congested_time[i] = fftime[i] * (
                    1.1 - (alpha[i]*(link_flows[i] / capacity[i]))) / (
                    1.1 - (link_flows[i] / capacity[i]))
        else:
            congested_time[i] = fftime[i]


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void dinrets_cython(
    double[:] deltaresult,
    double[:] link_flows,
    double [:] capacity,
    double [:] fftime,
    double[:] alpha,
    double [:] beta,
    int cores
) noexcept:
    cdef long long i
    cdef long long l = deltaresult.shape[0]

    for i in prange(l, nogil=True, num_threads=cores):
        if link_flows[i] > 0:
            if link_flows[i] > capacity[i]:
                deltaresult[i] = fftime[i] * (
                    (-20)*(alpha[i]-1.1)*link_flows[i]) / (
                    pow(capacity[i], 2))
            else:
                deltaresult[i] = fftime[i] * (
                    (-110)*(alpha[i]-1)*capacity[i]) / (
                    pow((11*capacity[i])-(10*link_flows[i]), 2))

        else:
            deltaresult[i] = fftime[i]



import random
from sqlite3 import Connection
from typing import Optional

from shapely.geometry import Point, MultiPolygon

from .network.connector_creation import connector_creation
from .network.safe_class import SafeClass


class Zone(SafeClass):
    """Single zone object that can be queried and manipulated in memory"""

    def __init__(self, dataset: dict, zoning):
        self.geometry = MultiPolygon()
        self.zone_id = -1
        super().__init__(dataset, zoning.project)
        self.__zoning = zoning
        self.__new = dataset["geometry"] is None
        self.__network_links = zoning.network.links
        self.__network_nodes = zoning.network.nodes

    def delete(self):
        """Removes the zone from the database"""
        with self.project.db_connection as conn:
            conn.execute(f'DELETE FROM zones where zone_id="{self.zone_id}"')
        self.__zoning._remove_zone(self.zone_id)
        del self

    def save(self):
        """Saves/Updates the zone data to the database"""

        if self.zone_id != self.__original__["zone_id"]:
            raise ValueError("One cannot change the zone_id")

        with self.project.db_connection as conn:
            if conn.execute(f'select count(*) from zones where zone_id="{self.zone_id}"').fetchone()[0] == 0:
                data = [self.zone_id, self.geometry.wkb]
                conn.execute("Insert into zones (zone_id, geometry) values(?, ST_Multi(GeomFromWKB(?, 4326)))", data)

            for key, value in self.__dict__.items():
                if key != "zone_id" and key in self.__original__:
                    v_old = self.__original__.get(key, None)
                    if value != v_old and value is not None:
                        self.__original__[key] = value
                        if key == "geometry":
                            sql = "update 'zones' set geometry=ST_Multi(GeomFromWKB(?, 4326)) where zone_id=?"
                            conn.execute(sql, [value.wkb, self.zone_id])
                        else:
                            conn.execute(f"update 'zones' set '{key}'=? where zone_id=?", [value, self.zone_id])

    def add_centroid(self, point: Point, robust=True) -> None:
        """Adds a centroid to the network file

        :Arguments:
            **point** (:obj:`Point`): Shapely Point corresponding to the desired centroid position.
            If None, uses the geometric center of the zone

            **robust** (:obj:`Bool`, *Optional*): Moves the centroid location around to avoid node conflict.
            Defaults to ``True``.
        """

        # This is VERY small in real-world terms (between zero and 11cm)
        shift = 0.000001

        with self.project.db_connection as conn:
            if conn.execute("select count(*) from nodes where node_id=?", [self.zone_id]).fetchone()[0] > 0:
                self.project.logger.warning("Centroid already exists. Failed to create it")
                return

            sql = "INSERT into nodes (node_id, is_centroid, geometry) VALUES(?,1,GeomFromWKB(?, ?));"

            if point is None:
                point = self.geometry.centroid

            if robust:
                check_sql = """SELECT count(*) FROM nodes
                                 WHERE  nodes.geometry = GeomFromWKB(?, 4326) AND
                              nodes.ROWID IN (
                               SELECT ROWID FROM SpatialIndex WHERE f_table_name = 'nodes' AND
                               search_frame = GeomFromWKB(?, 4326))
                           """

                test_list = conn.execute(check_sql, [point.wkb, point.wkb]).fetchone()
                while sum(test_list):
                    test_list = conn.execute(check_sql, [point.wkb, point.wkb]).fetchone()
                    point = Point(point.x + random.random() * shift, point.y + random.random() * shift)

            data = [self.zone_id, point.wkb, self.__srid__]
            conn.execute(sql, data)

    def connect_mode(
        self, mode_id: str, link_types="", connectors=1, conn: Optional[Connection] = None, limit_to_zone=True
    ) -> None:
        """Adds centroid connectors for the desired mode to the network file

        Centroid connectors are created by connecting the zone centroid to one or more nodes selected from
        all those that satisfy the mode and link_types criteria and are inside the zone.

        The selection of the nodes that will be connected is done simply by searching for the node closest to the
        zone centroid, or the N closest nodes to the centroid.

        If fewer candidates than required connectors are found, all candidates are connected.

        :Arguments:
            **mode_id** (:obj:`str`): Mode ID we are trying to connect

            **link_types** (:obj:`str`, *Optional*): String with all the link type IDs that can be considered.
            eg: yCdR. Defaults to ALL link types

            **connectors** (:obj:`int`, *Optional*): Number of connectors to add. Defaults to 1

            **conn** (:obj:`sqlite3.Connection`, *Optional*): Connection to the database.

            **limit_to_zone** (:obj:`bool`): Limits the search for nodes inside the zone. Defaults to ``True``.
        """

        area = self.geometry if limit_to_zone else None
        connector_creation(
            zone_id=self.zone_id,
            mode_id=mode_id,
            link_types=link_types,
            connectors=connectors,
            proj_nodes=self.project.network.nodes.data,
            proj_links=self.project.network.links.data,
            network=self.project.network,
            conn=conn,
            delimiting_area=area,
        )

    def disconnect_mode(self, mode_id: str) -> None:
        """Removes centroid connectors for the desired mode from the network file

        :Arguments:
            **mode_id** (:obj:`str`): Mode ID we are trying to disconnect from this zone
        """

        with self.project.db_connection as conn:
            data = [self.zone_id, mode_id]
            row_count = conn.execute("Delete from links where a_node=? and modes=?", data).rowcount

            data = [mode_id, self.zone_id, mode_id]
            sql = 'Update links set modes = replace(modes, ?, "") where a_node=? and instr(modes,?) > 0'
            row_count += conn.execute(sql, data).rowcount

            if row_count:
                self.project.logger.warning(
                    f"Deleted {row_count} connectors for mode {mode_id} for zone {self.zone_id}"
                )
            else:
                self.project.warning("No centroid connectors for this mode")



import multiprocessing as mp
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
from aequilibrae.paths.AoN import sum_axis1, assign_link_loads

from aequilibrae.matrix import AequilibraeMatrix
from aequilibrae.parameters import Parameters
from aequilibrae.paths.graph import Graph, TransitGraph, GraphBase, _get_graph_to_network_mapping

"""
TO-DO:
1. Make the writing to SQL faster by disabling all checks before the actual writing
"""


class AssignmentResultsBase(ABC):
    """Assignment results base class for traffic and transit assignments."""

    def __init__(self):
        self.link_loads = np.array([])  # The actual results for assignment
        self.no_path = None  # The list os paths
        self.num_skims = 0  # number of skims that will be computed. Depends on the setting of the graph provided
        p = Parameters().parameters["system"]["cpus"]
        if not isinstance(p, int):
            p = 0
        self.set_cores(p)

        self.nodes = -1
        self.zones = -1
        self.links = -1

        self.lids = None

    @abstractmethod
    def prepare(self, graph: GraphBase, matrix: AequilibraeMatrix) -> None:
        pass

    @abstractmethod
    def reset(self) -> None:
        pass

    def set_cores(self, cores: int) -> None:
        """
        Sets number of cores (threads) to be used in computation

        Value of zero sets number of threads to all available in the system, while negative values indicate the number
        of threads to be left out of the computational effort.

        Resulting number of cores will be adjusted to a minimum of zero or the maximum available in the system if the
        inputs result in values outside those limits

        :Arguments:
            **cores** (:obj:`int`): Number of cores to be used in computation
        """

        if not isinstance(cores, int):
            raise ValueError("Number of cores needs to be an integer")

        if cores < 0:
            self.cores = max(1, mp.cpu_count() + cores)
        elif cores == 0:
            self.cores = mp.cpu_count()
        elif cores > 0:
            cores = min(mp.cpu_count(), cores)
            if self.cores != cores:
                self.cores = cores
        if self.link_loads.shape[0]:
            self.__redim()


class AssignmentResults(AssignmentResultsBase):
    """
    Assignment result holder for a single :obj:`TrafficClass` with multiple user classes
    """

    def __init__(self):
        super().__init__()
        self.compact_link_loads = np.array([])  # Results for assignment on simplified graph
        self.compact_total_link_loads = np.array([])  # Results for all user classes summed on simplified graph
        self.crosswalk = np.array([])  # crosswalk between compact graph link IDs and actual link IDs
        self.skims = AequilibraeMatrix()  # The array of skims
        self.total_link_loads = np.array([])  # The result of the assignment for all user classes summed

        self.compact_links = -1
        self.compact_nodes = -1

        self.direcs = None

        self.classes = {"number": 1, "names": ["flow"]}

        self._selected_links = {}
        self.select_link_od = None
        self.select_link_loading = {}

        self._graph_id = None
        self.__float_type = None
        self.__integer_type = None

        # save path files. Need extra metadata for file paths
        self.save_path_file = False
        self.path_file_dir = None
        self.write_feather = True  # we use feather as default, parquet is slower but with better compression

    # In case we want to do by hand, we can prepare each method individually
    def prepare(self, graph: Graph, matrix: AequilibraeMatrix) -> None:
        """
        Prepares the object with dimensions corresponding to the assignment matrix and graph objects

        :Arguments:
            **graph** (:obj:`Graph`): Needs to have been set with number of centroids and list of skims (if any)

            **matrix** (:obj:`AequilibraeMatrix`): Matrix properly set for computation with
            ``matrix.computational_view(:obj:`list`)``
        """

        self.__float_type = graph.default_types("float")
        self.__integer_type = graph.default_types("int")

        if matrix.view_names is None:
            raise ValueError("Please set the matrix_procedures computational view")
        self.classes["number"] = 1
        if len(matrix.matrix_view.shape) > 2:
            self.classes["number"] = matrix.matrix_view.shape[2]
        self.classes["names"] = matrix.view_names

        if graph is None:
            raise ValueError("Please provide a graph")
        self.compact_nodes = graph.compact_num_nodes
        self.compact_links = graph.compact_num_links

        self.nodes = graph.num_nodes
        self.zones = graph.num_zones
        self.centroids = graph.centroids
        self.links = graph.num_links
        self.num_skims = len(graph.skim_fields)
        self.skim_names = list(graph.skim_fields)
        self.lids = graph.graph.link_id.values
        self.direcs = graph.graph.direction.values
        self.crosswalk = np.zeros(graph.graph.shape[0], self.__integer_type)
        self.crosswalk[graph.graph.__supernet_id__.values] = graph.graph.__compressed_id__.values
        self._graph_ids = graph.graph.__supernet_id__.values
        self._graph_compressed_ids = graph.graph.__compressed_id__.values
        self.__redim()
        self._graph_id = graph._id

        if self._selected_links:
            self.select_link_od = AequilibraeMatrix()
            self.select_link_od.create_empty(
                memory_only=True,
                zones=matrix.zones,
                matrix_names=list(self._selected_links.keys()),
                index_names=matrix.index_names,
            )

            self.select_link_loading = {}
            # Combine each set of selected links into one large matrix that can be parsed into Cython
            # Each row corresponds a link set, and the equivalent rows in temp_sl_od_matrix and temp_sl_link_loading
            # Correspond to that set
            self.select_links = np.full(
                (len(self._selected_links), max([len(x) for x in self._selected_links.values()])),
                -1,
                dtype=graph.default_types("int"),
            )

            sl_idx = {}
            for i, (name, arr) in enumerate(self._selected_links.items()):
                sl_idx[name] = i
                # Filling select_links array with linksets. Note the default value is -1, which is used as a placeholder
                # It also denotes when the given row has no more selected links, since Cython cannot handle
                # Multidimensional arrays where each row has different lengths
                self.select_links[i][: len(arr)] = arr
                # Correctly sets the dimensions for the final output matrices
                self.select_link_od.matrix[name] = np.zeros(
                    (graph.num_zones, graph.num_zones, self.classes["number"]),
                    dtype=graph.default_types("float"),
                )
                self.select_link_loading[name] = np.zeros(
                    (graph.compact_num_links, self.classes["number"]),
                    dtype=graph.default_types("float"),
                )

            # Overwrites previous arrays on assignment results level with the index to access that array in Cython
            self._selected_links = sl_idx

    def reset(self) -> None:
        """
        Resets object to prepared and pre-computation state
        """
        if self.num_skims > 0:
            self.skims.matrices.fill(0)
        if self.link_loads is not None:
            self.no_path.fill(0)
            self.link_loads.fill(0)
            self.total_link_loads.fill(0)
            self.compact_link_loads.fill(0)
            self.compact_total_link_loads.fill(0)
        else:
            raise ValueError("Exception: Assignment results object was not yet prepared/initialized")

    def __redim(self):
        self.compact_link_loads = np.zeros((self.compact_links + 1, self.classes["number"]), self.__float_type)
        self.compact_total_link_loads = np.zeros(self.compact_links, self.__float_type)

        self.link_loads = np.zeros((self.links, self.classes["number"]), self.__float_type)
        self.total_link_loads = np.zeros(self.links, self.__float_type)
        self.no_path = np.zeros((self.zones, self.zones), dtype=self.__integer_type)

        if self.num_skims > 0:
            self.skims = AequilibraeMatrix()

            self.skims.create_empty(file_name=self.skims.random_name(), zones=self.zones, matrix_names=self.skim_names)
            self.skims.index[:] = self.centroids[:]
            self.skims.computational_view()
            if len(self.skims.matrix_view.shape[:]) == 2:
                self.skims.matrix_view = self.skims.matrix_view.reshape((self.zones, self.zones, 1))
        else:
            self.skims = AequilibraeMatrix()
            self.skims.matrix_view = np.array((1, 1, 1))

        self.reset()

    def total_flows(self) -> None:
        """
        Totals all link flows for this class into a single link load

        Results are placed into *total_link_loads* class member
        """
        sum_axis1(self.total_link_loads, self.link_loads, self.cores)

    def get_graph_to_network_mapping(self):
        return _get_graph_to_network_mapping(self.lids, self.direcs)

    def get_load_results(self) -> pd.DataFrame:
        """
        Translates the assignment results from the graph format into the network format

        :Returns:
            **dataset** (:obj:`pd.DataFrame`): Pandas DataFrame data with the traffic class assignment results
        """

        # Get a mapping from the compressed graph to/from the network graph
        m = self.get_graph_to_network_mapping()

        recs = np.unique(self.lids).shape[0]

        # Link flows
        link_flows = self.link_loads[self._graph_ids, :]
        aux = {}
        for i, n in enumerate(self.classes["names"]):
            # Directional Flows
            aux[n + "_ab"] = np.zeros(recs, self.__float_type)
            aux[n + "_ab"][m.network_ab_idx] = np.nan_to_num(link_flows[m.graph_ab_idx, i])

            aux[n + "_ba"] = np.zeros(recs, self.__float_type)
            aux[n + "_ba"][m.network_ba_idx] = np.nan_to_num(link_flows[m.graph_ba_idx, i])

            # Tot Flow
            aux[n + "_tot"] = np.nan_to_num(aux[n + "_ab"]) + np.nan_to_num(aux[n + "_ba"])

        return pd.DataFrame(aux, index=np.unique(self.lids))

    def get_sl_results(self) -> pd.DataFrame:
        # Set up the name for each column. Each set of select links has a column for ab, ba, total flows
        # for each subclass contained in the TrafficClass
        fields = [
            e
            for name in self._selected_links.keys()
            for n in self.classes["names"]
            for e in [f"{name}_{n}_ab", f"{name}_{n}_ba", f"{name}_{n}_tot"]
        ]

        res = pd.DataFrame([], columns=fields, index=np.unique(self.lids))

        m = self.get_graph_to_network_mapping()
        for name in self._selected_links.keys():
            # Link flows initialised
            link_flows = np.full((self.links, self.classes["number"]), np.nan)
            # maps link flows from the compressed graph to the uncompressed graph
            assign_link_loads(link_flows, self.select_link_loading[name], self._graph_compressed_ids, self.cores)
            for i, n in enumerate(self.classes["names"]):
                # Directional Flows
                res[f"{name}_{n}_ab"].values[m.network_ab_idx] = link_flows[m.graph_ab_idx, i]
                res[f"{name}_{n}_ba"].values[m.network_ba_idx] = link_flows[m.graph_ba_idx, i]

                # Tot Flow
                res[f"{name}_{n}_tot"] = np.nansum(res[[f"{name}_{n}_ab", f"{name}_{n}_ba"]].to_numpy(), axis=1)

        return res


class TransitAssignmentResults(AssignmentResultsBase):
    """
    Assignment result holder for a single :obj:`Transit`
    """

    def __init__(self):
        super().__init__()

        self.link_loads = np.array([])

    def prepare(self, graph: TransitGraph, matrix: AequilibraeMatrix) -> None:
        """
        Prepares the object with dimensions corresponding to the assignment matrix and graph objects

        :Arguments:
            **graph** (:obj:`TransitGraph`): Needs to have been set with number of centroids

            **matrix** (:obj:`AequilibraeMatrix`): Matrix properly set for computation with
            ``matrix.computational_view(:obj:`list`)``
        """
        self.reset()
        self.nodes = graph.num_nodes
        self.zones = graph.num_zones
        self.centroids = graph.centroids
        self.links = graph.num_links
        self.lids = graph.graph.link_id.values

    def reset(self) -> None:
        """
        Resets object to prepared and pre-computation state
        """

        # Since all memory for the assignment is managed by the HyperpathGenerating
        # object we don't need to do much here
        self.link_loads.fill(0)

    def get_load_results(self) -> pd.DataFrame:
        """
        Translates the assignment results from the graph format into the network format

        :Returns:
            **dataset** (:obj:`pd.DataFrame`): DataFrame data with the transit class assignment results
        """
        if not self.link_loads.shape[0]:
            raise ValueError("Transit assignment has not been executed yet")

        return pd.DataFrame({"volume": self.link_loads}, index=self.lids)



"""
An implementation of Spiess and Florian's hyperpath generating algorithm.

reference: Spiess, H. and Florian, M. (1989). Optimal strategies: A new
assignment model for transit networks. Transportation Research Part B 23(2),
83-102.
"""

cimport cython
import numpy as np
cimport numpy as cnp

from cython.parallel import parallel, prange, threadid
from libc.stdlib cimport malloc, calloc, free

ctypedef cnp.float64_t DATATYPE_t
DATATYPE_PY = np.float64

cdef DATATYPE_t DATATYPE_INF = <DATATYPE_t>np.finfo(dtype=DATATYPE_PY).max
DATATYPE_INF_PY = DATATYPE_INF

# infinite frequency is defined here numerically
# this must be a very large number depending on the precision on the computation
# INF_FREQ << DATATYPE_INF
cdef DATATYPE_t INF_FREQ = 1.0e+20
INF_FREQ_PY = INF_FREQ

# smallest frequency
# WARNING: this must be small but not too small
# 1 / MIN_FREQ << DATATYPE_INF
cdef DATATYPE_t MIN_FREQ
MIN_FREQ = 1.0 / INF_FREQ
MIN_FREQ_PY = MIN_FREQ

# a very small time interval
cdef DATATYPE_t A_VERY_SMALL_TIME_INTERVAL
A_VERY_SMALL_TIME_INTERVAL = 1.0e+08 * MIN_FREQ
A_VERY_SMALL_TIME_INTERVAL_PY = A_VERY_SMALL_TIME_INTERVAL

cdef extern from "stdlib.h":
    ctypedef void const_void "const void"
    void qsort(void *base, int nmemb, int size, int(*compar)(const_void *, const_void *)) noexcept nogil

cdef struct IndexedElement:
    size_t index
    DATATYPE_t value

cdef int _compare(const_void *a, const_void *b) noexcept:
    cdef DATATYPE_t v = (<IndexedElement*> a).value-(<IndexedElement*> b).value
    if v < 0:
        return -1
    else:
        return 1

include 'pq_4ary_heap.pyx'  # priority queue


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.embedsignature(False)
@cython.initializedcheck(False)
cdef void _coo_tocsc_uint32(
    cnp.uint32_t [::1] Ai,
    cnp.uint32_t [::1] Aj,
    cnp.uint32_t [::1] Ax,
    cnp.uint32_t [::1] Bp,
    cnp.uint32_t [::1] Bi,
    cnp.uint32_t [::1] Bx
) noexcept nogil:

    cdef:
        size_t i, col, dest
        size_t n_vert = <size_t>(Bp.shape[0] - 1)
        size_t n_edge = <size_t>Bi.shape[0]
        cnp.uint32_t temp, cumsum, last

    for i in range(n_edge):
        Bp[<size_t>Aj[i]] += 1

    cumsum = 0
    for i in range(n_vert):
        temp = Bp[i]
        Bp[i] = cumsum
        cumsum += temp
    Bp[<size_t>n_vert] = <cnp.uint32_t>n_edge

    for i in range(n_edge):
        col  = <size_t>Aj[i]
        dest = <size_t>Bp[col]
        Bi[dest] = Ai[i]
        Bx[dest] = Ax[i]
        Bp[col] += 1

    last = 0
    for i in range(n_vert + 1):
        temp = Bp[i]
        Bp[i] = last
        last = temp


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.embedsignature(False)
@cython.initializedcheck(False)
cdef void argsort(cnp.float64_t *data, cnp.uint32_t *order, size_t n) noexcept nogil:
    """
    Wrapper of the C function qsort
    source: https://github.com/jcrudy/cython-argsort/tree/master/cyargsort
    """
    cdef size_t i

    # Allocate index tracking array.
    cdef IndexedElement *order_struct = <IndexedElement *> malloc(n * sizeof(IndexedElement))

    # Copy data into index tracking array.
    for i in range(n):
        order_struct[i].index = i
        order_struct[i].value = data[i]

    # Sort index tracking array.
    qsort(<void *> order_struct, n, sizeof(IndexedElement), _compare)

    # Copy indices from index tracking array to output array.
    for i in range(n):
        order[i] = <cnp.uint32_t>order_struct[i].index

    # Free index tracking array.
    free(order_struct)


cpdef convert_graph_to_csc_uint32(edges, tail, head, data, vertex_count):
    """
    Convert an edge dataframe in COO format into CSC format.

    The data vector is of uint32 type.

    Parameters
    ----------
    edges : pandas.core.frame.DataFrame
        The edges dataframe.
    tail : str
        The column name in the edges dataframe for the tail vertex index.
    head : str
        The column name in the edges dataframe for the head vertex index.
    data : str
        The column name in the edges dataframe for the int edge attribute.
    vertex_count : int
        The vertex count in the given network edges.

    Returns
    -------
    tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]
    """

    rs_indptr = np.zeros(
        vertex_count + 1, dtype=np.uint32
    )  # make sure it is filled with zeros
    edge_count = len(edges)
    rs_indices = np.empty(edge_count, dtype=np.uint32)
    rs_data = np.empty(edge_count, dtype=np.uint32)

    _coo_tocsc_uint32(
        edges[tail].values.astype(np.uint32),
        edges[head].values.astype(np.uint32),
        edges[data].values.astype(np.uint32),
        rs_indptr,
        rs_indices,
        rs_data,
    )

    return rs_indptr, rs_indices, rs_data


# Both boundscheck(False) and initializedcheck(False) are required for this function to operate,
# without them the threads appear to enter a deadlock due to the shared edge_volume array. However
# the indexing on that array and its slices should never collide.
#
# Optionally return the u_i_vec, however this is only possible when using a single destination.
# The caller is responsible for managing that memory.
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.embedsignature(False)
@cython.initializedcheck(False)
cdef void compute_SF_in_parallel(
    cnp.uint32_t[::1] indptr_view,  # column csc format
    cnp.uint32_t[::1] edge_idx_view,
    cnp.float64_t[::1] trav_time_view,
    cnp.float64_t[::1] freq_view,
    cnp.uint32_t[::1] tail_view,
    cnp.uint32_t[::1] head_view,
    cnp.uint32_t[:] d_vert_ids_view,  # destination vertices
    # in public_transport.run case, same as d_vert_ids_view in .assign are the unique destination values
    cnp.uint32_t[:] destination_vertex_indices_view,
    cnp.uint32_t[::1] o_vert_ids_view,  # origin vertices
    cnp.float64_t[::1] demand_vls_view,  # volume
    cnp.float64_t[::1] edge_volume_view,
    size_t vertex_count,
    size_t edge_count,
    int num_threads,
    cnp.float64_t[:, ::1] skim_col_view,
    cnp.float64_t[::1] u_i_vec_view,
    cnp.float64_t[:, :, ::1] skim_matrix,
    cnp.uint32_t[::1] centroids,
    cnp.uint32_t[::1] centroids_idx_pos,
    bint skimming,
    bint is_travel_time,
    size_t n_skim_cols,
) noexcept nogil:
    # Thread local variables are prefixed by "thread", anything else should be considered shared and thus read only
    cdef:
        cnp.uint32_t *thread_demand_origins
        cnp.float64_t *thread_demand_values
        cnp.float64_t *thread_edge_volume
        size_t demand_size

        cnp.float64_t *thread_u_i_vec
        cnp.float64_t *thread_f_i_vec
        cnp.float64_t *thread_u_j_c_a_vec
        cnp.float64_t *thread_v_i_vec
        cnp.uint8_t *thread_h_a_vec
        cnp.uint32_t *thread_edge_indices

        cnp.float64_t *thread_skim_i_vec
        cnp.float64_t *thread_skim_j_vec

        # This is a shared buffer, all threads will write into separate slices depending on their threadid.
        # When writing all threads must increment!
        cnp.float64_t *edge_volume = <cnp.float64_t *> calloc(num_threads, sizeof(cnp.float64_t) * edge_count)

        cnp.float64_t *u_i_vec_out = <cnp.float64_t *> malloc(num_threads * sizeof(cnp.float64_t) * vertex_count)

        cnp.float64_t *skim_i_vec_out = <cnp.float64_t *> calloc(
            num_threads, sizeof(cnp.float64_t) * vertex_count * n_skim_cols)

        int i  # openmp on windows requires iterator variable have signed type
        size_t k, j, destination_vertex_index

    with parallel(num_threads=min(num_threads, centroids.shape[0] if skimming else d_vert_ids_view.shape[0])):
        thread_demand_origins = <cnp.uint32_t  *> malloc(sizeof(cnp.uint32_t)  * d_vert_ids_view.shape[0])
        thread_demand_values  = <cnp.float64_t *> malloc(sizeof(cnp.float64_t) * d_vert_ids_view.shape[0])
        # Here we take out thread local slice of the shared buffer, each thread is assigned a unique id so
        # we can safely read and write without collisions.
        thread_edge_volume  = &edge_volume[threadid() * edge_count]
        thread_u_i_vec = &u_i_vec_out[threadid() * vertex_count]

        thread_skim_i_vec = &skim_i_vec_out[threadid() * vertex_count * n_skim_cols]

        thread_f_i_vec      = <cnp.float64_t *> malloc(sizeof(cnp.float64_t) * vertex_count)
        thread_u_j_c_a_vec  = <cnp.float64_t *> malloc(sizeof(cnp.float64_t) * edge_count)
        thread_v_i_vec      = <cnp.float64_t *> malloc(sizeof(cnp.float64_t) * vertex_count)
        thread_h_a_vec      = <cnp.uint8_t   *> malloc(sizeof(cnp.uint8_t)   * edge_count)
        thread_edge_indices = <cnp.uint32_t  *> malloc(sizeof(cnp.uint32_t)  * edge_count)

        thread_skim_j_vec = <cnp.float64_t *> calloc(edge_count, sizeof(cnp.float64_t) * n_skim_cols)

        for i in prange(destination_vertex_indices_view.shape[0]):
            destination_vertex_index = destination_vertex_indices_view[i]

            demand_size = 0
            for j in range(d_vert_ids_view.shape[0]):
                if d_vert_ids_view[j] == destination_vertex_index:
                    thread_demand_origins[demand_size] = o_vert_ids_view[j]
                    thread_demand_values[demand_size] = demand_vls_view[j]
                    # demand_size += 1 is not allowed as cython believes this is a reduction
                    demand_size = demand_size + 1

            # S&F
            compute_SF_in(
                indptr_view,
                edge_idx_view,
                trav_time_view,
                freq_view,
                tail_view,
                head_view,
                thread_demand_origins,
                thread_demand_values,
                demand_size,
                thread_edge_volume,
                thread_u_i_vec,
                thread_f_i_vec,
                thread_u_j_c_a_vec,
                thread_v_i_vec,
                thread_h_a_vec,
                thread_edge_indices,
                vertex_count,
                destination_vertex_index,
                skim_col_view,
                thread_skim_i_vec,
                thread_skim_j_vec,
                n_skim_cols,
                skim_matrix,
                centroids,
                centroids_idx_pos,
                skimming,
                is_travel_time
            )

        free(thread_demand_origins)
        free(thread_demand_values)
        free(thread_f_i_vec)
        free(thread_u_j_c_a_vec)
        free(thread_v_i_vec)
        free(thread_h_a_vec)
        free(thread_edge_indices)
        free(thread_skim_j_vec)

    # Accumulate results into output buffer. This could parallelised over the output indexes but
    # the lose of spacial locality may not be worth it.
    for i in range(num_threads):
        for j in range(edge_count):
            edge_volume_view[j] += edge_volume[i * edge_count + j]

    if d_vert_ids_view.shape[0] == 1:
        for k in range(vertex_count):
            u_i_vec_view[k] = u_i_vec_out[k]

    free(u_i_vec_out)
    free(skim_i_vec_out)
    free(edge_volume)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.embedsignature(False)
@cython.initializedcheck(False)
cdef void compute_SF_in(
    cnp.uint32_t[::1] csc_indptr,
    cnp.uint32_t[::1] csc_edge_idx,
    cnp.float64_t[::1] c_a_vec,  # travel time
    cnp.float64_t[::1] f_a_vec,  # freq
    cnp.uint32_t[::1] tail_indices,
    cnp.uint32_t[::1] head_indices,
    cnp.uint32_t *demand_indices,
    cnp.float64_t *demand_values,
    size_t demand_size,
    cnp.float64_t *v_a_vec,
    cnp.float64_t *u_i_vec,
    cnp.float64_t *f_i_vec,
    cnp.float64_t *u_j_c_a_vec,
    cnp.float64_t *v_i_vec,
    cnp.uint8_t *h_a_vec,
    cnp.uint32_t *edge_indices,
    size_t vertex_count,
    int dest_vert_index,
    cnp.float64_t[:, ::1] skim_col_vec,
    cnp.float64_t *skim_i_vec,
    cnp.float64_t *skim_j_vec,
    size_t n_skim_cols,
    cnp.float64_t[:, :, ::1] skim_matrix,
    cnp.uint32_t[::1] centroids,
    cnp.uint32_t[::1] centroids_idx_pos,
    bint skimming,
    bint is_travel_time,
) noexcept nogil:

    cdef:
        size_t edge_count = <size_t>tail_indices.shape[0]
        DATATYPE_t u_r, u_i
        size_t i, j, h_a_count
        cnp.uint32_t vert_idx
        int cent_dest

    # initialization
    for i in range(vertex_count):
        u_i_vec[i] = DATATYPE_INF
        f_i_vec[i] = 0.0
        v_i_vec[i] = 0.0
    u_i_vec[<size_t>dest_vert_index] = 0.0

    for i in range(vertex_count * n_skim_cols):
        skim_i_vec[i] = 0.0

    for i in range(edge_count):
        # v_a_vec[i] = 0.0
        u_j_c_a_vec[i] = DATATYPE_INF
        h_a_vec[i] = 0

    # first pass #
    # ---------- #
    _SF_in_first_pass_full(
        csc_indptr,
        csc_edge_idx,
        c_a_vec,
        f_a_vec,
        tail_indices,
        u_i_vec,
        f_i_vec,
        u_j_c_a_vec,
        h_a_vec,
        dest_vert_index,
        skim_col_vec,
        skim_i_vec,
        skim_j_vec,
        n_skim_cols,
        vertex_count
    )

    # store skimming results
    cent_dest = centroids_idx_pos[dest_vert_index]

    if skimming:
        if is_travel_time:
            for i in range(centroids.shape[0]):
                skim_matrix[cent_dest][i][0] = u_i_vec[centroids[i]]

            if n_skim_cols > 0:
                for j in range(<size_t>n_skim_cols):
                    for i in range(centroids.shape[0]):
                        skim_matrix[cent_dest][i][j+1] = skim_i_vec[centroids[i] + j * vertex_count]

        else:
            for j in range(<size_t>n_skim_cols):
                for i in range(centroids.shape[0]):
                    skim_matrix[cent_dest][i][j] = skim_i_vec[centroids[i] + j * vertex_count]

    # second pass #
    # ----------- #

    # demand is loaded into all the origin vertices
    # also we compute the min travel time from all the origin vertices
    u_r = DATATYPE_INF
    for i in range(demand_size):
        vert_idx = demand_indices[i]  # origin vertex id
        v_i_vec[<size_t>vert_idx] = demand_values[i]
        u_i = u_i_vec[<size_t>vert_idx]
        if u_i < u_r:
            u_r = u_i

    # if the destination can be reached from any of the origins
    if u_r < DATATYPE_INF:

        # make sure f_i values are not zero
        for i in range(<size_t>vertex_count):
            if f_i_vec[i] < MIN_FREQ:
                f_i_vec[i] = MIN_FREQ

        h_a_count = 0
        for i in range(<size_t>edge_count):
            u_j_c_a_vec[i] *= -1.0
            h_a_count += <size_t>h_a_vec[i]

        # Sort the links with decreasing order of u_j + c_a.
        # Because the sort function sorts in increasing order, we sort a
        # transformed array, multiplied by -1, and set the items
        # corresponding to edges that are not in the hyperpath to a
        # positive value (they will be at the end of the sorted array).
        # The items corresponding to edges that are in the hyperpath have a
        # negative transformed value.
        for i in range(<size_t>edge_count):
            if h_a_vec[i] == 0:  # if the edge is not on the hyperpath
                u_j_c_a_vec[i] = 1.0

        argsort(u_j_c_a_vec, edge_indices, edge_count)

        _SF_in_second_pass(
            edge_indices,
            tail_indices,
            head_indices,
            v_i_vec,
            v_a_vec,
            f_i_vec,
            f_a_vec,
            h_a_count
        )


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.embedsignature(False)
@cython.cdivision(True)
@cython.initializedcheck(False)
cdef void _SF_in_first_pass_full(
    cnp.uint32_t[::1] csc_indptr,
    cnp.uint32_t[::1] csc_edge_idx,
    cnp.float64_t[::1] c_a_vec,  # travel time
    cnp.float64_t[::1] f_a_vec,  # freq
    cnp.uint32_t[::1] tail_indices,
    cnp.float64_t *u_i_vec,
    cnp.float64_t *f_i_vec,
    cnp.float64_t *u_j_c_a_vec,
    cnp.uint8_t *h_a_vec,
    int dest_vert_index,
    cnp.float64_t[:, ::1] skim_col_vec,
    cnp.float64_t *skim_i_vec,
    cnp.float64_t *skim_j_vec,
    size_t n_skim_cols,
    size_t vertex_count,
) noexcept nogil:
    """All vertices are visited."""

    cdef:
        int edge_count = tail_indices.shape[0]
        PriorityQueue pqueue
        ElementState edge_state
        size_t i, j, edge_idx, tail_vert_idx
        DATATYPE_t u_j_c_a, u_i, f_i, beta, u_i_new, f_a
        DATATYPE_t skim_i, skim_j, skim_i_new, beta_skim

        cnp.float64_t *skim_i_new_vec = <cnp.float64_t *> malloc(n_skim_cols * sizeof(cnp.float64_t))
        cnp.float64_t *skim_j_new_vec = <cnp.float64_t *> malloc(n_skim_cols * sizeof(cnp.float64_t))

    # initialization of the heap elements
    # all nodes have INFINITY key and NOT_IN_HEAP state
    init_heap(&pqueue, <size_t>edge_count)

    # only the incoming edges of the target vertex are inserted into the
    # priority queue
    for i in range(<size_t>csc_indptr[<size_t>dest_vert_index], <size_t>csc_indptr[<size_t>(dest_vert_index + 1)]):
        edge_idx = csc_edge_idx[i]
        insert(&pqueue, edge_idx, c_a_vec[edge_idx])
        u_j_c_a_vec[edge_idx] = c_a_vec[edge_idx]
        for j in range(<size_t>n_skim_cols):
            skim_j_vec[edge_idx + j * edge_count] = skim_col_vec[edge_idx][j]

    # first pass
    while pqueue.size > 0:

        edge_idx = extract_min(&pqueue)
        u_j_c_a = pqueue.Elements[edge_idx].key
        tail_vert_idx = <size_t>tail_indices[edge_idx]
        u_i = u_i_vec[tail_vert_idx]

        if u_i >= u_j_c_a:

            f_i = f_i_vec[tail_vert_idx]

            # compute the beta coefficient
            if (u_i < DATATYPE_INF) | (f_i > 0.0):
                beta = f_i * u_i

            else:
                beta = 1.0

            # update u_i
            f_a = f_a_vec[edge_idx]

            u_i_new = (beta + f_a * u_j_c_a) / (f_i + f_a)
            u_i_vec[tail_vert_idx] = u_i_new

            for j in range(<size_t>n_skim_cols):

                skim_i = skim_i_vec[tail_vert_idx + j * vertex_count]
                skim_j = skim_j_vec[edge_idx + j * edge_count]

                if (u_i < DATATYPE_INF) | (f_i > 0.0):
                    beta_skim = f_i * skim_i
                else:
                    beta_skim = 0.0

                skim_i_new = (beta_skim + f_a * skim_j) / (f_i + f_a)

                skim_i_vec[tail_vert_idx  + j * vertex_count] = skim_i_new
                skim_i_new_vec[j] = skim_i_new

            # update f_i
            f_i_vec[tail_vert_idx] = f_i + f_a

            # add the edge to hyperpath
            h_a_vec[edge_idx] = 1

        else:

            u_i_new = u_i
            for j in range(<size_t>n_skim_cols):
                skim_i_new = skim_i_vec[tail_vert_idx + j * vertex_count]
                skim_i_new_vec[j] = skim_i_new

        # loop on incoming edges
        for i in range(<size_t>csc_indptr[tail_vert_idx], <size_t>csc_indptr[tail_vert_idx + 1]):

            edge_idx = csc_edge_idx[i]
            edge_state = pqueue.Elements[edge_idx].state

            if edge_state != SCANNED:

                # u_j of current edge = u_i of outgoing edge
                u_j_c_a = u_i_new + c_a_vec[edge_idx]

                for j in range(<size_t>n_skim_cols):
                    skim_j_new_vec[j] = skim_i_new_vec[j] + skim_col_vec[edge_idx][j]

                if edge_state == NOT_IN_HEAP:

                    insert(&pqueue, edge_idx, u_j_c_a)
                    u_j_c_a_vec[edge_idx] = u_j_c_a
                    for j in range(<size_t>n_skim_cols):
                        skim_j_vec[edge_idx + j * edge_count] = skim_j_new_vec[j]

                elif (pqueue.Elements[edge_idx].key > u_j_c_a):

                    decrease_key(&pqueue, edge_idx, u_j_c_a)
                    u_j_c_a_vec[edge_idx] = u_j_c_a
                    for j in range(<size_t>n_skim_cols):
                        skim_j_vec[edge_idx + j * edge_count] = skim_j_new_vec[j]

    free_heap(&pqueue)
    free(skim_i_new_vec)
    free(skim_j_new_vec)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.embedsignature(False)
@cython.cdivision(True)
@cython.initializedcheck(False)
cdef void _SF_in_second_pass(
    cnp.uint32_t *edge_indices,
    cnp.uint32_t[::1] tail_indices,
    cnp.uint32_t[::1] head_indices,
    cnp.float64_t *v_i_vec,
    cnp.float64_t *v_a_vec,
    cnp.float64_t *f_i_vec,
    cnp.float64_t[::1] f_a_vec,
    size_t h_a_count
) noexcept nogil:

    cdef:
        size_t i, edge_idx, vert_idx
        DATATYPE_t v_i, f_i, f_a, v_a_new

    for i in range(h_a_count):

        edge_idx = <size_t>edge_indices[i]
        vert_idx = <size_t>tail_indices[edge_idx]

        v_i = v_i_vec[vert_idx]
        f_i = f_i_vec[vert_idx]
        f_a = f_a_vec[edge_idx]

        # update v_a
        v_a_new = v_i * f_a / f_i
        v_a_vec[edge_idx] += v_a_new
        v_i_vec[<size_t>head_indices[edge_idx]] += v_a_new



from typing import Union

from aequilibrae.project.network.mode import Mode
from .safe_class import SafeClass


class Link(SafeClass):
    """A Link object represents a single record in the *links* table

    .. code-block:: python

        >>> project = create_example(project_path)

        >>> all_links = project.network.links

        # Let's get a mode to work with
        >>> modes = project.network.modes
        >>> car_mode = modes.get('c')

        # We can just get one link in specific
        >>> link1 = all_links.get(3)
        >>> link2 = all_links.get(17)

        # We can find out which fields exist for the links
        >>> which_fields_do_we_have = link1.data_fields()

        # And edit each one like this
        >>> link1.lanes_ab = 3
        >>> link1.lanes_ba = 2

        # we can drop a mode from the link
        >>> link1.drop_mode(car_mode)  # or link1.drop_mode('c')

        # we can add a mode to the link
        >>> link2.add_mode(car_mode)  # or link2.add_mode('c')

        # Or set all modes at once
        >>> link2.set_modes('cbtw')

        # We can just save the link
        >>> link1.save()
        >>> link2.save()
    """

    def __init__(self, dataset, project):
        super().__init__(dataset, project)
        self.__fields = list(dataset.keys())

        self.__new = dataset["geometry"] is None
        self.__stil_exists = True
        self._table = "links"

    def delete(self):
        """Deletes link from database"""
        with self.project.db_connection as conn:
            conn.execute(f'DELETE FROM links where link_id="{self.link_id}"')
        self.__stil_exists = False

    def save(self, conn=None):
        """Saves link to database"""

        data, sql = self._save_new_with_geometry() if self.__new else self.__save_existing_link()

        if data:
            with conn or self.project.db_connection as conn:
                conn.execute(sql, data)

        self.__new = False

        for key in self.__original__.keys():
            self.__original__[key] = self.__dict__[key]

    def set_modes(self, modes: str):
        """Sets the modes acceptable for this link

        :Arguments:
            **modes** (:obj:`str`): string with all mode_ids to be assigned to this link
        """

        if not isinstance(modes, str):
            raise ValueError("Modes field needs to be a string")
        if modes == "":
            raise ValueError("Modes field needs to have at least one mode")

        self.__dict__["modes"] = modes

    def add_mode(self, mode: Union[str, Mode]):
        """Adds a new mode to this link

        Raises a warning if mode is already allowed on the link, and fails if mode does not exist

        :Arguments:
            **mode_id** (:obj:`str` or `Mode`): Mode_id of the mode or mode object to be added to the link
        """
        mode_id = self.__validate(mode)

        if mode_id in self.modes:
            self._logger.warning("Mode already active for this link")
            return

        self.__dict__["modes"] += mode_id

    def drop_mode(self, mode: Union[str, Mode]):
        """Removes a mode from this link

        Raises a warning if mode is already NOT allowed on the link, and fails if mode does not exist

        :Arguments:
            **mode_id** (:obj:`str` or `Mode`): Mode_id of the mode or mode object to be removed from the link
        """

        mode_id = self.__validate(mode)

        if mode_id not in self.modes:
            self._logger.warning("Mode already inactive for this link")
            return

        if len(self.modes) == 1:
            raise ValueError("Link needs to have at least one mode")

        self.__dict__["modes"] = self.modes.replace(mode_id, "")

    def data_fields(self) -> list:
        """lists all data fields for the link, as available in the database

        :Returns:
            **data fields** (:obj:`list`): list of all fields available for editing
        """

        return list(self.__original__.keys())

    def _exists(self):
        return self.__stil_exists

    def __validate(self, mode: Union[str, Mode]) -> str:
        if isinstance(mode, Mode):
            mode_id = mode.mode_id
        elif isinstance(mode, str):
            if len(mode) > 1:
                raise ValueError("A mode_id is a single character")
            mode_id = mode
        else:
            raise TypeError("You should provide a mode id (string) or a Mode object")
        return mode_id

    def __save_existing_link(self):
        data = []
        if self.link_id != self.__original__["link_id"]:
            raise ValueError("One cannot change the link_id")

        txts = []
        for key, val in self.__dict__.items():
            if key not in self.__original__:
                continue
            if val != self.__original__[key]:
                if key == "geometry" and val is not None:
                    data.extend([val.wkb, self.__srid__])
                    txts.append("geometry=GeomFromWKB(?, ?)")
                else:
                    data.append(val)
                    txts.append(f'"{key}"=?')

        if not data:
            self._logger.warning(f"Nothing to update for link {self.link_id}")
            return [], ""

        txts = ",".join(txts) + " where link_id=?"
        data.append(self.link_id)
        sql = f"Update Links set {txts}"
        return data, sql

    def __setattr__(self, instance, value) -> None:
        if instance not in self.__dict__ and instance[:1] != "_":
            raise AttributeError(f'"{instance}" is not a valid attribute for a link')
        if instance == "modes":
            self.set_modes(value)
        elif instance == "a_node":
            raise AttributeError("Setting a_node is not allowed")
        elif instance == "b_node":
            raise AttributeError("Setting b_node is not allowed")
        elif instance == "link_id":
            raise ValueError("Changing a link_id is not supported. Create a new one and delete this")
        else:
            self.__dict__[instance] = value



from libcpp.vector cimport vector
from libcpp.utility cimport pair
from libcpp.unordered_set cimport unordered_set
from libcpp.unordered_map cimport unordered_map
from libcpp cimport bool
from libcpp.memory cimport unique_ptr

# std::linear_congruential_engine is not available in the Cython libcpp.random shim. We'll import it ourselves
# from libcpp.random cimport minstd_rand
from libc.stdint cimport *

cdef extern from "<random>" namespace "std" nogil:
    cdef cppclass random_device:
        ctypedef uint_fast32_t result_type
        random_device() except +
        result_type operator()() except +

    cdef cppclass minstd_rand:
        ctypedef uint_fast32_t result_type
        minstd_rand() except +
        minstd_rand(result_type seed) except +
        result_type operator()() except +
        result_type min() except +
        result_type max() except +
        void discard(size_t z) except +
        void seed(result_type seed) except +

# std::shuffle is not available in the Cython libcpp.algorithm shim. We'll import it ourselves
# from libcpp.algorithm cimport shuffle
cdef extern from "<algorithm>" namespace "std" nogil:
    void shuffle[RandomIt, URBG](RandomIt first, RandomIt last, URBG&& g) except +

# std::make_pair is not available in the Cython libcpp.utilities shim. We'll import it ourselves based on the C++11 til
# C++14 definition because C++14 makes this signature weird. See https://github.com/cython/cython/issues/2706
cdef extern from "<utility>" namespace "std" nogil:
    pair[T, U] make_pair[T, U](T&& t, U&& u)

# To define our own hashing functions we have to write a little C++. The string is inlined directly into
# route_choice.cpp To make Cython aware of our hash types we also must declare them with the right signatures
#
# OrderedVectorPointerHasher: This hash function is for hashing the routes, thus it should be order *DEPENDENT*.
# Potential performance improvements may come from https://en.wikipedia.org/wiki/Universal_hashing#Hashing_vectors
#
# UnorderedSetPointerHasher: This hash function is for hashing the banned route sets, thus it should be order
# *INDEPENDENT*.  Potential performance improvements may come from:
# Mark N. Wegman, J.Lawrence Carter,
# New hash functions and their use in authentication and set equality
# https://doi.org/10.1016/0022-0000(81)90033-7
#
# PointerDereferenceEqualTo: Because we are storing and hashing the pointers to objects to avoid unnecessary copies we
# must define our own comparator to resolve hash collisions. Without this equality operator the bare pointers are
# compared.
cdef extern from * nogil:
    """
    // Source: https://stackoverflow.com/a/72073933
    // License: CC BY-SA 4.0 Deed, https://creativecommons.org/licenses/by-sa/4.0/
    struct OrderedVectorPointerHasher {
        size_t operator()(const std::vector<long long> *V) const {
            size_t seed = V->size();
            long long x;
            for(auto &i : *V) {
                x = ((i >> 16) ^ i) * 0x45d9f3b;
                x = ((x >> 16) ^ x) * 0x45d9f3b;
                x = (x >> 16) ^ x;
                seed ^= x + 0x9e3779b9 + (seed << 6) + (seed >> 2);
            }
            return seed;
        }
    };

    // Source: https://stackoverflow.com/a/1537189
    // License: CC BY-SA 2.5 Deed, https://creativecommons.org/licenses/by-sa/2.5/
    struct UnorderedSetPointerHasher {
        size_t operator()(const std::unordered_set<long long> *S) const {
            size_t hash = 1;
            for(auto &i : *S) {
                hash *= 1779033703 + 2 * i;
            }
            return hash / 2;
        }
    };

    template<class T>
    struct PointerDereferenceEqualTo {
        bool operator()(const T& lhs, const T& rhs) const {
            return *lhs == *rhs;
        }
    };
    """
    cppclass OrderedVectorPointerHasher:
        size_t operator()(const vector[long long] *V) const

    cppclass UnorderedSetPointerHasher:
        size_t operator()(const unordered_set[long long] *S) const

    cppclass PointerDereferenceEqualTo[T]:
        bool operator()(const T& lhs, const T& rhs) const

# For typing convenience, the types names are getting long
ctypedef unordered_set[
    vector[long long] *,
    OrderedVectorPointerHasher,
    PointerDereferenceEqualTo[vector[long long] *]
] RouteSet_t
ctypedef unordered_set[
    unordered_set[long long] *,
    UnorderedSetPointerHasher,
    PointerDereferenceEqualTo[unordered_set[long long] *]
] LinkSet_t
ctypedef vector[pair[unordered_set[long long] *, vector[long long] *]] RouteMap_t

ctypedef vector[unique_ptr[vector[long long]]] RouteVec_t

# A (known 2016) bug in the Cython compiler means it incorrectly parses the following type when used in a cdef
# https://github.com/cython/cython/issues/534
ctypedef vector[bool]* vector_bool_ptr



import math
from sqlite3 import Connection
from shapely.geometry import LineString, Point
from shapely.ops import substring

from aequilibrae.transit.constants import Constants, TRANSIT_LINK_RANGE


class Link:
    """Transit link element.

    * transit_link (:obj:`int`): ID of the transit link (updated when inserted in the database)
    * from_stop (:obj:`str`): Origin of the transit connection
    * to_stop (:obj:`str`): Destination of the transit connection
    * pair (:obj:`str`): Identifier of the stop pair as FROM_ID##TO_ID. For identification only
    * geo (:obj:`LineString`): Geometry of the transit link as direct connection between stops
    * length (:obj:`float`): Link length measured directly from the geometry object
    * type (:obj:`int`): Route type (mode) for this transit link
    * srid (:obj:`int`): srid of our working database
    """

    def __init__(self, srid) -> None:
        """
        :Arguments:
            *srid* (:obj:`int`): srid of our working database
        """

        self.__dict__["from_stop"] = ""
        self.__dict__["to_stop"] = ""
        self.__dict__["pattern_id"] = -1
        self.__dict__["geo"] = None
        self.transit_link = -1
        self.pattern_id = -1
        self.from_stop = ""
        self.to_stop = ""
        self.key = "##"
        self.geo = None  # type: LineString
        self.length = -1
        self.type = -1
        self.seq = None
        self.srid = srid

    def __setattr__(self, key, value):
        self.__dict__[key] = value
        self.__dict__["key"] = f"{self.from_stop}##{self.to_stop}##{self.pattern_id}"
        if self.geo is not None:
            self.__dict__["length"] = self.geo.length * math.pi * 6371000 / 180

    def build_geo(self, from_point: Point, to_point: Point, gtfs_shape: LineString, previous_end: Point):
        """Builds link geometry."""
        geo = LineString([from_point, to_point])
        if gtfs_shape is not None:
            fpos = gtfs_shape.project(from_point)
            tpos = gtfs_shape.project(to_point)
            if fpos < tpos:
                geo2 = substring(gtfs_shape, fpos, tpos)
                geo = geo if geo2.length / geo.length > 2 else geo2

        self.geo = geo if geo.touches(previous_end) else LineString([previous_end] + list(geo.coords))

    def save_to_database(self, conn: Connection, commit=True) -> None:
        """Saves Transit link to the database"""

        data = [
            self.pattern_id,
            self.seq,
            self.transit_link,
            self.from_stop,
            self.to_stop,
            self.length,
            self.geo.wkb,
            self.srid,
        ]
        sql = """insert into route_links (pattern_id, seq, transit_link, from_stop, to_stop, distance, geometry)
                                            values (?, ?, ?, ?, ?, ?, GeomFromWKB(?, ?));"""
        conn.execute(sql, data)
        if commit:
            conn.commit()

    def get_link_id(self):
        c = Constants()
        self.transit_link = c.transit_links.get(1, TRANSIT_LINK_RANGE) + 1
        c.transit_links[1] = self.transit_link



from os import PathLike
from pathlib import Path
from typing import Union

import geopandas as gpd
import pandas as pd

from aequilibrae.utils.db_utils import read_and_close
from aequilibrae.utils.get_table import get_geo_table


class DataLoader:
    def __init__(self, path_to_file: PathLike, table_name: str):
        self.__pth_file = Path(path_to_file)
        self.table_name = table_name

    def load_table(self) -> Union[gpd.GeoDataFrame, pd.DataFrame]:
        with read_and_close(self.__pth_file, spatial=True) as conn:
            return get_geo_table(self.table_name, conn)



import itertools
import warnings
import logging
import pathlib
import socket
import sqlite3
from datetime import datetime
from typing import List, Optional, Tuple, Union, Dict
from collections.abc import Hashable
from os import path
from uuid import uuid4
from functools import cached_property

import numpy as np
import pandas as pd
import openmatrix as omx
import scipy
from aequilibrae.context import get_active_project
from aequilibrae.matrix import AequilibraeMatrix
from aequilibrae.paths.graph import Graph, _get_graph_to_network_mapping
from aequilibrae.paths.cython.route_choice_set import RouteChoiceSet
from aequilibrae.paths.cython.route_choice_set_results import RouteChoiceSetResults
from aequilibrae.matrix.coo_demand import GeneralisedCOODemand
from aequilibrae.utils.db_utils import commit_and_close


class RouteChoice:
    all_algorithms = ["bfsle", "lp", "link-penalisation", "link-penalization"]

    default_parameters = {
        "generic": {
            "seed": 0,
            "max_routes": 0,
            "max_depth": 0,
            "max_misses": 100,
            "penalty": 1.01,
            "cutoff_prob": 0.0,
            "beta": 1.0,
            "store_results": True,
        },
        "link-penalisation": {},
        "bfsle": {"penalty": 1.0},
    }

    demand_index_names = ["origin id", "destination id"]

    def __init__(self, graph: Graph, project=None):
        self.parameters = {}
        self.procedure_id = None
        self.procedure_date = None

        proj = project or get_active_project(must_exist=False)
        self.project = proj

        self.logger = proj.logger if proj else logging.getLogger("aequilibrae")

        self.cores: int = 0
        self.graph = graph
        self.demand = self.__init_demand()

        self.sl_compact_link_loads: Optional[Dict[str, np.array]] = None
        self.sl_link_loads: Optional[Dict[str, np.array]] = None

        self.where: Optional[pathlib.Path] = None
        self.index_name = "route_choice_sl_index"

        self._config = {}
        self._selected_links = {}
        self.sl_link_loading = True

    @cached_property
    def __rc(self) -> RouteChoiceSet:
        return RouteChoiceSet(self.graph)

    def __init_demand(self):
        d = GeneralisedCOODemand(
            *self.demand_index_names, self.graph.nodes_to_indices, shape=(self.graph.num_zones, self.graph.num_zones)
        )
        return d

    def set_choice_set_generation(self, /, algorithm: str = None, **kwargs) -> None:
        """
        Chooses the assignment algorithm and set its parameters.

        Options for algorithm are 'bfsle' for breadth first search with link removal, or
        'link-penalisation'/'link-penalization'. 'lp' is also accepted as an alternative to
        'link-penalisation'. If ``algorithm`` is ``None``, none will be set, but the parameters
        will be updated. This is useful when assigning from path-files.

        BFSLE implementation based on "Route choice sets for very high-resolution data" by
        Nadine Rieser-Schüssler, Michael Balmer & Kay W. Axhausen (2013).
        `DOI: 10.1080/18128602.2012.671383 <https://doi.org/10.1080/18128602.2012.671383>`_.

        Setting the parameters for the route choice:

        * ``seed`` is a BFSLE specific parameters.

        * Although not required, setting ``max_depth`` or ``max_misses``, is strongly
          recommended to prevent runaway algorithms.

        * ``max_misses`` is the maximum amount of duplicate routes found per OD pair. If a set
          of routes is returned in a case where ``max_misses`` is exceeded, the number of routes
          may be fewer than ``max_routes``. Assumes a default value of 100.

        * When using **BFSLE** ``max_depth`` corresponds to the maximum height of the graph.
          It's value is largely dependent on the size of the paths within the network. For
          very small networks a value of 10 is a recommended starting point. For large networks
          a good starting value is 5. Increase the value until the number of desired routes is
          being consistently returned. If a set of routes is returned in a case where
          ``max_depth`` is exceeded, the number of routes may be fewer than ``max_routes``.

        * When using **LP**, ``max_depth`` corresponds to the maximum number of iterations
          performed. While not enforced, it should be higher than ``max_routes``. It's value is
          dependent on the magnitude of the cost field, specifically if it's related to the log
          base ``penalty`` of the ratio of costs between two alternative routes. If a set of
          routes is returned in a case where ``max_depth`` is exceeded, the number of routes
          may be fewer than ``max_routes``.

        * Additionally BFSLE has the option to incorporate link penalisation. Every link in all
          routes found at a depth are penalised with the penalty factor for the next depth. So at
          a depth of 0 no links are penalised nor removed. At depth 1, all links found at depth 0
          are penalised, then the links marked for removal are removed. All links in the routes
          found at depth 1 are then penalised for the next depth. The penalisation compounds.
          Set ``penalty=1.0`` to disable.

        * When performing an assignment, ``cutoff_prob`` can be provided to exclude routes from
          the path-sized logit model. The ``cutoff_prob`` is used to compute an inverse binary
          logit and obtain a max difference in utilities. If a paths total cost is greater than
          the minimum cost path in the route set plus the max difference, the route is excluded
          from the PSL calculations. The route is still returned, but with a probability of 0.0.

        * The ``cutoff_prob`` should be in the range :math:`[0, 1]`. It is then rescaled
          internally to :math:`[0.5, 1]` as probabilities below 0.5 produce negative differences
          in utilities because the choice is between two routes only, one of which is the
          shortest path. A higher ``cutoff_prob`` includes less routes. A value of 1.0 will only
          include the minimum cost route. A value of 0.0 includes all routes.

        :Arguments:
            **algorithm** (:obj:`str`): Algorithm to be used

            **kwargs** (:obj:`dict`): Dictionary with all parameters for the algorithm
        """
        if algorithm is not None:
            algo_dict = {i: i for i in self.all_algorithms}
            algo_dict["lp"] = "link-penalisation"
            algo_dict["link-penalization"] = "link-penalisation"

            sentinel = object()
            algorithm = algo_dict.get(algorithm.lower(), sentinel)

            if algorithm is sentinel:
                raise AttributeError(
                    f"Assignment algorithm not available. Choose from: {','.join(self.all_algorithms)}"
                )

            defaults = self.default_parameters["generic"] | self.default_parameters[algorithm]
            for key in kwargs.keys():
                if key not in defaults:
                    raise ValueError(f"Invalid parameter '{key}' provided for algorithm `{algorithm}`")
        else:
            defaults = self.default_parameters["generic"]
            for key in kwargs.keys():
                if key not in defaults:
                    raise ValueError(f"Invalid or non-generic parameter '{key}' provided")

        self.algorithm = algorithm
        self._config["Algorithm"] = algorithm

        self.parameters = defaults | kwargs

    def set_cores(self, cores: int) -> None:
        """Allows one to set the number of cores to be used

            Inherited from :obj:`AssignmentResultsBase`

        :Arguments:
            **cores** (:obj:`int`): Number of CPU cores to use
        """
        self.cores = cores

    def set_save_routes(self, where: Optional[str] = None) -> None:
        """
        Set save path for route choice results. Provide ``None`` to disable.

        .. warning:

            Enabling route saving will disable in memory results. Viewing the results will read
            the results from disk first.

        :Arguments:
            **save_it** (:obj:`bool`): Boolean to indicate whether routes should be saved
        """

        if where is not None:
            where = pathlib.Path(where)
            if not where.exists():
                raise ValueError(f"Path does not exist `{where}`")
        self.where = where

    def add_demand(self, demand, fill: float = 0.0):
        """
        Add demand DataFrame or matrix for the assignment.

        :Arguments:
            **demand** (:obj:`Union[pd.DataFrame, AequilibraeMatrix]`): Demand to add to
            assignment. If the supplied demand is a DataFrame, it should have a 2-level
            MultiIndex of Origin and Destination node IDs. If an AequilibraE Matrix is
            supplied node IDs will be inferred from the index. Demand values should be either
            ``float32``s or ``float64``s.

            **fill** (:obj:`float`): Value to fill any ``NaN``s with.
        """
        if isinstance(demand, pd.DataFrame):
            self.demand.add_df(demand, fill=fill)
        elif isinstance(demand, AequilibraeMatrix):
            self.demand.add_matrix(demand, fill=fill)
        else:
            raise TypeError(f"unknown argument type '{(type(demand).__name__)}'")

    def prepare(self, nodes: Union[List[int], List[Tuple[int, int]], None] = None) -> None:
        """
        Prepare OD pairs for batch computation.

        :Arguments:
            **nodes** (:obj:`Union[list[int], list[tuple[int, int]]]`): List of node IDs to
            operate on. If a 1D list is provided, OD pairs are taken to be all pair
            permutations of the list. If a list of pairs is provided OD pairs are taken as is.
            All node IDs must be present in the compressed graph. To make a node ID always appear
            in the compressed graph add it as a centroid. Duplicates will be dropped on execution.
            If ``None`` is provided, all OD pairs with non-zero flows will be used.
        """
        if nodes is not None and not self.demand.no_demand():
            raise ValueError("provide either `nodes` or set a `demand` matrix, not both")
        elif nodes is None:
            return
        elif len(nodes) == 0:
            raise ValueError("`nodes` list-like empty.")

        self.demand = self.__init_demand()
        df = pd.DataFrame()
        if all(
            isinstance(pair, tuple)
            and len(pair) == 2
            and isinstance(pair[0], (int, np.integer))
            and isinstance(pair[1], (int, np.integer))
            for pair in nodes
        ):
            df.index = pd.MultiIndex.from_tuples(nodes, name=self.demand_index_names)
        elif len(nodes) > 1 and all(isinstance(x, (int, np.unsignedinteger)) for x in nodes):
            df.index = pd.MultiIndex.from_tuples(itertools.permutations(nodes, r=2), names=self.demand_index_names)
        else:
            raise ValueError(f"{type(nodes)} or {type(nodes[0])} for not valid types for the `prepare` method")

        self.demand.add_df(df)

    def execute_single(self, origin: int, destination: int, demand: float = 0.0) -> List[Tuple[int]]:
        """
        Generate route choice sets between origin and destination, potentially performing an
        assignment.

        Does not require preparation.

        Node IDs must be present in the compressed graph. To make a node ID always appear in
        the compressed graph add it as a centroid.

        :Arguments:
            **origin** (:obj:`int`): Origin node ID.

            **destination** (:obj:`int`): Destination node ID.

            **demand** (:obj:`float`): If provided an assignment will be performed with this demand.

        :Returns:
            **route set** (:obj:`List[Tuple[int]]`): A list of routes as tuples of link IDs.
        """
        self.procedure_id = uuid4().hex
        self.procedure_date = str(datetime.today())

        return self.__rc.run(
            origin,
            destination,
            self.demand.shape,
            demand=demand,
            bfsle=self.algorithm == "bfsle",
            path_size_logit=bool(demand),
            cores=self.cores,
            where=str(self.where) if self.where is not None else None,
            sl_link_loading=self.sl_link_loading,
            **self.parameters,
        )

    def execute(self, perform_assignment: bool = True) -> None:
        """
        Generate route choice sets between the previously supplied nodes, potentially performing
        an assignment.

        To access results see ``RouteChoice.get_results()``.

        :Arguments:
            **perform_assignment** (:obj:`bool`): Whether or not to perform an assignment.
            Defaults to ``False``.
        """
        if self.demand.df.index.empty:
            logging.warning("There is no demand or pairs of OD pairs to compute Route choice for.")
            return

        self.procedure_date = str(datetime.today())

        self.__rc.batched(
            self.demand,
            self._selected_links,
            bfsle=self.algorithm == "bfsle",
            path_size_logit=perform_assignment,
            cores=self.cores,
            where=str(self.where) if self.where is not None else None,
            sl_link_loading=self.sl_link_loading,
            **self.parameters,
        )

    def execute_from_path_files(self, path_files: Union[pathlib.Path, str], recompute_psl: bool = False) -> None:
        """
        Perform an assignment from an existing set of path-files.

        This method expects the path-files to be written by the ``self.save_path_files()`` method,
        however any parquet hive dataset with the correct structure is accepted. This allows the
        use of AequilibraE's path-sized logit, link loading, select link analysis, and assignment
        while using externally generated routes.
        """

        # Read the dataset schema and make sure it conforms to what we want
        df = RouteChoiceSetResults.read_dataset(path_files)
        required_fields = ["origin id", "destination id", "route set"] + [] if recompute_psl else ["probability"]
        schema = {
            "destination id": "uint32",
            "route set": "object",
            "cost": "float64",
            "mask": "bool",
            "path overlap": "float64",
            "probability": "float64",
            "origin id": "uint32",
        }
        dtypes = df.dtypes

        try:
            for field in required_fields:
                if schema[field] != dtypes.loc[field]:
                    raise TypeError(
                        f"schema of required field '{field}' does not match. "
                        f"Expected {schema[field]}, "
                        f"found {dtypes.loc[field]}"
                    )
        except KeyError as e:
            raise KeyError(f"Column '{field}' does not exist in the path-files") from e

        self.execute_from_pandas(df=df, recompute_psl=recompute_psl)

    def execute_from_pandas(self, df: pd.DataFrame, recompute_psl: bool = False) -> None:
        """
        Perform an assignment using route sets from a Pandas DataFrame.

        Requires the DataFrame contains the ``origin id``, ``destination id`` and ``route set``
        columns. The route sets must be a list of links IDs stored as integers with the direction
        encoded as the sign.  Additionally, when ``recompute_psl`` is ``False``, the
        ``probability`` column must also be present.

        When ``recompute_psl`` is ``True``, the path-sized logit is recomputed for each route with
        respect to the graphs current cost field and the ``beta`` and ``cutoff_prob`` parameters.

        All origin and destination IDs within the DataFrame must exist within the demand matrix.

        All link IDs and directions must exist within the graph. Links must also be present within
        the compressed graph.

        If ``recompute_psl`` is ``False`` the table returned from ``self.get_results()`` will have
        all zeros for the cost and path overlap fields, and all True for the mask field. If
        ``recompute_psl`` is ``True`` these fields will be recalculated as required.
        """

        required_columns = ["origin id", "destination id", "route set"] + [] if recompute_psl else ["probability"]
        for col in required_columns:
            if col not in df.columns:
                raise ValueError(f"provided DataFrame is missing required column '{col}'")

        self.__rc.assign_from_df(
            self.graph.graph,
            df,
            self.demand,
            select_links=self._selected_links,
            recompute_psl=recompute_psl,
            sl_link_loading=self.sl_link_loading,
            store_results=self.parameters["store_results"],
            beta=self.parameters["beta"],
            cutoff_prob=self.parameters["cutoff_prob"],
        )

    def info(self) -> dict:
        """Returns information for the transit assignment procedure

        Dictionary contains keys:

        * Algorithm,
        * Matrix totals
        * Computer name
        * Procedure ID
        * Parameters
        * Select links

        The classes key is also a dictionary with all the user classes per transit class and
        their respective matrix totals.

        :Returns:
            **info** (:obj:`dict`): Dictionary with summary information
        """

        matrix_totals = self.demand.df.sum().to_dict()

        info = {
            "Algorithm": self.algorithm,
            "Matrix totals": matrix_totals,
            "Computer name": socket.gethostname(),
            "Procedure ID": self.procedure_id,
            "Parameters": self.parameters,
            "Select links": self._selected_links,
        }
        return info

    def log_specification(self):
        self.logger.info("Route Choice specification")
        self.logger.info(self._config)

    def get_results(self) -> pd.DataFrame:
        """
        Returns the results of the route choice procedure

        Returns a table of OD pairs to lists of link IDs for each OD pair provided (as columns).  Represents paths from
        ``origin`` to ``destination``. When the link id in the route set is positive it represents the ab direction,
        while negative represents the ba direction.

        :Returns:
            **results** (:obj:`pd.DataFrame`): Table with the results of the route choice procedure
        """
        if self.where is None:
            results = self.__rc.get_results()
        else:
            results = RouteChoiceSetResults.read_dataset(self.where)

        return results

    def save_path_files(self, where: Optional[pathlib.Path] = None):
        """
        Save path-files to the directory specific.

        Files will be saved as a parquet hive dataset partitioned by the origin ID. Existing path-files will not be
        removed to allow incremental route choice set generation.

        :Arguments:
            **where** (:obj:`Optional[pathlib.Path]`): Directory to save the dataset to.
        """
        where = where if where is not None else self.where
        if where is None:
            raise ValueError("either the 'where' argument or 'self.where' property must not None")

        self.__rc.write_path_files(where)

    def get_load_results(self) -> pd.DataFrame:
        """
        Translates the link loading results from the graph format into the network format.

        :Returns:
            **dataset** (:obj:`Union[Tuple[pd.DataFrame, pd.DataFrame], pd.DataFrame]`): A tuple
            of link loading results as DataFrames. Columns are the matrix name concatenated
            direction.
        """

        if self.demand.no_demand():
            warnings.warn("No demand was provided. To perform link loading add a demand matrix or data frame")
            return pd.DataFrame([])

        ll = self.__rc.get_link_loading(cores=self.cores)
        ll = {(k,): v for k, v in ll.items()}

        # Create a data store with a row for each uncompressed link
        m = _get_graph_to_network_mapping(self.graph.graph.link_id.values, self.graph.graph.direction.values)
        lids = np.unique(self.graph.graph.link_id.values)
        df = self.__link_loads_to_df(m, lids, ll)

        return df

    def __link_loads_to_df(self, mapping, lids, link_loads):
        df = pd.DataFrame(
            {"link_id": lids} | {(*k, dir): np.zeros(lids.shape) for k in link_loads.keys() for dir in ["ab", "ba"]}
        )
        added_dfs = []
        for k, v in link_loads.items():
            # Directional Flows
            df.iloc[mapping.network_ab_idx, df.columns.get_loc((*k, "ab"))] = np.nan_to_num(v[mapping.graph_ab_idx])
            df.iloc[mapping.network_ba_idx, df.columns.get_loc((*k, "ba"))] = np.nan_to_num(v[mapping.graph_ba_idx])

            # Tot Flow
            added_dfs.append(pd.DataFrame({(*k, "tot"): df[(*k, "ab")] + df[(*k, "ba")]}))

        df = pd.concat([df] + added_dfs, axis=1).set_index("link_id")
        df.columns = pd.MultiIndex.from_tuples(df.columns)
        return df.sort_index()

    def set_select_links(
        self, links: Dict[Hashable, List[Union[Tuple[int, int], List[Tuple[int, int]]]]], link_loading=True
    ):
        """
        Set the selected links. Checks if the links and directions are valid. Supports **OR** and
        **AND** sets of links.

        Dictionary values should be a list of either a single ``(link_id, direction)`` tuple or a
        list of ``(link_id, dirirection)``.

        The elements of the first list represent the **AND** sets, together they are OR'ed. If any
        of these sets is satisfied the link are loaded as appropriate.

        The **AND** sets are comprised of either a single ``(link_id, direction)`` tuple or a list
        of ``(link_id, direction)``. The single tuple represents an **AND** set with a single
        element.

        All links and directions in an **AND** set must appear in any order within a route for it
        to be considered satisfied.

        Supply ``links=None`` to disable select link analysis.

        :Arguments:
            **links** (:obj:`Union[None, Dict[Hashable, List[Union[Tuple[int, int], List[Tuple[int, int]]]]]]`):
            Name of link set and link IDs and directions to be used in select link analysis.

            **link_loading** (:obj:`bool`): Enable select link loading. If disabled only OD matrix
            results are available.

        """
        self._selected_links = {}
        self.sl_link_loading = link_loading

        if links is None:
            del self._config["select_links"]
            return

        max_id = self.graph.compact_graph.id.max() + 1

        for name, link_set in links.items():
            normalised_link_set = []
            for link_ids in link_set:
                if isinstance(link_ids, tuple) and len(link_ids) == 2 and link_ids[1] == 0:
                    warnings.warn(
                        f"Adding both directions of a link ({link_ids[0]}) to a single AND set is likely "
                        f"unintentional. Replacing with {(link_ids[0], -1)} OR {(link_ids[0], 1)}"
                    )
                    normalised_link_set.append((link_ids[0], -1))
                    normalised_link_set.append((link_ids[0], 1))
                else:
                    normalised_link_set.append(link_ids)

            or_set = set()
            for link_ids in normalised_link_set:
                and_set = set()
                for link, dir in link_ids:
                    if dir == 0:
                        query = (self.graph.graph["link_id"] == link) & (
                            (self.graph.graph["direction"] == -1) | (self.graph.graph["direction"] == 1)
                        )
                    else:
                        query = (self.graph.graph["link_id"] == link) & (self.graph.graph["direction"] == dir)

                    if not query.any():
                        raise ValueError(f"link_id or direction {(link, dir)} is not present within graph.")

                    for comp_id in self.graph.graph[query]["__compressed_id__"].values:
                        # Check for duplicate compressed link ids in the current link set
                        if comp_id == max_id:
                            raise ValueError(
                                f"link ID {link} and direction {dir} is not present in compressed graph. "
                                "It may have been removed during dead-end removal."
                            )
                        elif comp_id in and_set:
                            warnings.warn(
                                "Two input links map to the same compressed link in the network"
                                f", removing superfluous link {link} and direction {dir} with compressed id {comp_id}"
                            )
                        else:
                            and_set.add(comp_id)

                or_set.add(frozenset(and_set))
            self._selected_links[name] = frozenset(or_set)
        self._config["select_links"] = str(links)

    def get_select_link_loading_results(self) -> pd.DataFrame:
        """
        Get the select link loading results.

        :Returns:
            **dataset** (:obj:`Tuple[pd.DataFrame, pd.DataFrame]`): Select link loading results
            as DataFrames. Columns are the matrix name concatenated with the select link set
            and direction.
        """

        if self.demand.no_demand():
            raise ValueError("no demand was provided. To perform link loading add a demand matrix or data frame")
        elif not self.sl_link_loading:
            raise ValueError("select link loading was disabled via `set_select_links(..., link_loading=False)`")

        sl_link_loads = {}
        for sl_name, sl_res in self.__rc.get_sl_link_loading().items():
            for demand_name, res in sl_res.items():
                sl_link_loads[demand_name, sl_name] = res

        # Create a data store with a row for each uncompressed link
        m = _get_graph_to_network_mapping(self.graph.graph.link_id.values, self.graph.graph.direction.values)
        lids = np.unique(self.graph.graph.link_id.values)
        df = self.__link_loads_to_df(m, lids, sl_link_loads)

        return df

    def get_select_link_od_matrix_results(self) -> Dict[str, Dict[str, scipy.sparse.coo_matrix]]:
        """
        Get the select link OD matrix results as a sparse matrix.

        :Returns:
            **select link OD matrix results** (:obj:`Dict[str, Dict[str, scipy.sparse.coo_matrix]]`):
            Returns a dict of select link set names to a dict of demand column names to a sparse OD
            matrix
        """

        if self.demand.no_demand():
            raise ValueError("No demand was provided. To perform link loading add a demand matrix or data frame")

        return self.__rc.get_sl_od_matrices()

    def __save_dataframe(self, df, method_name: str, description: str, table_name: str, report: dict, project) -> None:
        self.procedure_id = uuid4().hex
        data = [
            table_name,
            "select link",
            self.procedure_id,
            str(report),
            self.procedure_date,
            description,
        ]

        # sqlite3 context managers only commit, they don't close, oh well
        res_path = path.join(project.project_base_path, "results_database.sqlite")
        with commit_and_close(res_path, missing_ok=True) as conn:
            df.to_sql(table_name, conn, index=True)

        with self.project.db_connection as conn:
            conn.execute(
                """Insert into results(table_name, procedure, procedure_id, procedure_report, timestamp,
                                                description) Values(?,?,?,?,?,?)""",
                data,
            )

    def save_link_flows(self, table_name: str, project=None) -> None:
        """
        Saves the link link flows for all classes into the results database.

        :Arguments:
            **table_name** (:obj:`str`): Name of the table being inserted to.

            **project** (:obj:`Project`, `Optional`): Project we want to save the results to.
            Defaults to the active project
        """
        if not project:
            project = self.project or get_active_project()

        df = self.get_load_results()
        info = self.info()
        self.__save_dataframe(
            df,
            "Link loading",
            "Uncompressed link loading results",
            table_name + "_uncompressed",
            info,
            project=project,
        )

    def save_select_link_flows(self, table_name: str, project=None) -> None:
        """
        Saves the select link link flows for all classes into the results database. Additionally,
        it exports the OD matrices into OMX format.

        :Arguments:
            **table_name** (:obj:`str`): Name of the table being inserted to and the name of the
            OpenMatrix file used for OD matrices.

            **project** (:obj:`Project`, `Optional`): Project we want to save the results to.
            Defaults to the active project
        """
        if not project:
            project = self.project or get_active_project()

        u = self.get_select_link_loading_results()
        info = self.info()
        self.__save_dataframe(
            u,
            "Select link analysis",
            "Uncompressed select link analysis results",
            table_name + "_uncompressed",
            info,
            project=project,
        )

        path = (pathlib.Path(project.project_base_path) / "matrices" / table_name).with_suffix(".omx")
        for sl_name, v in self.get_select_link_od_matrix_results().items():
            for demand_name, mat in v.items():
                mat.to_disk(path, sl_name + "_" + demand_name)

        file = omx.open_file(path, "a")
        try:
            file.create_mapping(self.index_name, self.graph.centroids)
        finally:
            file.close()
